import ast
import numpy as np
import pandas as pd
from typing import Tuple

from nemo_bo.opt.objectives import CalculableObjective, RegressionObjective, ObjectivesList
from nemo_bo.opt.variables import (
    CategoricalVariableWithDescriptors,
    CategoricalVariableDiscreteValues,
    ContinuousVariable,
    VariablesList,
)


def extract_input_template(input_template_xlsx: str) -> Tuple[np.ndarray, np.ndarray, VariablesList, ObjectivesList]:
    """
    Function that can extract user-inputted variables and objectives information and data from the input template
    provided

    Parameters
    ----------
    input_template_xlsx: str
        Excel file for the input template containing the variables, objectives, and data

    Returns
    -------
    X: np.ndarry
        Data for the variables
    Y: np.ndarray
        Data for the objectives
    var_list: VariablesList
        Object that contains all of the variables information
    obj_list: VariablesList
        Object that contains all of the objectives information
    """
    input_template = pd.read_excel(input_template_xlsx, sheet_name=None, index_col=0)

    var_list = []
    for var_index, var_type in enumerate(input_template["Variables"].columns):
        if "ContinuousVariable" in var_type:
            var_list.append(
                ContinuousVariable(
                    name=input_template["Variables"].loc["name"][var_index],
                    lower_bound=input_template["Variables"].loc["lower_bound"][var_index],
                    upper_bound=input_template["Variables"].loc["upper_bound"][var_index],
                    transformer=input_template["Variables"].loc["transformer"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["transformer"][var_index])
                    else None,
                    units=input_template["Variables"].loc["units"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["units"][var_index])
                    else None,
                )
            )
        elif "CategoricalVariableDiscreteValues" in var_type:
            name = input_template["Variables"].loc["name"][var_index]
            var_list.append(
                CategoricalVariableDiscreteValues(
                    name=name,
                    categorical_levels=list(input_template[name].index),
                    lower_bound=input_template["Variables"].loc["lower_bound"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["lower_bound"][var_index])
                    else None,
                    upper_bound=input_template["Variables"].loc["upper_bound"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["upper_bound"][var_index])
                    else None,
                    transformer=input_template["Variables"].loc["transformer"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["transformer"][var_index])
                    else None,
                    units=input_template["Variables"].loc["units"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["units"][var_index])
                    else None,
                )
            )
        elif "CategoricalVariableWithDescriptors" in var_type:
            name = input_template["Variables"].loc["name"][var_index]
            var_list.append(
                CategoricalVariableWithDescriptors(
                    name=name,
                    categorical_levels=list(input_template[name].index),
                    descriptor_names=list(input_template[name].columns),
                    categorical_descriptors=input_template[name].values,
                    lower_bound=input_template["Variables"].loc["lower_bound"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["lower_bound"][var_index])
                    else None,
                    upper_bound=input_template["Variables"].loc["upper_bound"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["upper_bound"][var_index])
                    else None,
                    transformer=input_template["Variables"].loc["transformer"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["transformer"][var_index])
                    else None,
                    units=input_template["Variables"].loc["units"][var_index]
                    if not pd.isnull(input_template["Variables"].loc["units"][var_index])
                    else None,
                )
            )

    obj_list = []
    for obj_index, obj_type in enumerate(input_template["Objectives"].columns):
        if "RegressionObjective" in obj_type:
            obj_list.append(
                RegressionObjective(
                    name=input_template["Objectives"].loc["name"][obj_index],
                    obj_max_bool=input_template["Objectives"].loc["obj_max_bool"][obj_index],
                    lower_bound=input_template["Objectives"].loc["lower_bound"][obj_index],
                    upper_bound=input_template["Objectives"].loc["upper_bound"][obj_index],
                    transformer=input_template["Objectives"].loc["transformer"][obj_index]
                    if not pd.isnull(input_template["Objectives"].loc["transformer"][obj_index])
                    else None,
                    units=input_template["Objectives"].loc["units"][obj_index]
                    if not pd.isnull(input_template["Objectives"].loc["units"][obj_index])
                    else None,
                    predictor_type=ast.literal_eval(input_template["Objectives"].loc["predictor_type"][obj_index])
                    if not pd.isnull(input_template["Objectives"].loc["predictor_type"][obj_index])
                    else None,
                )
            )
        elif "CalculableObjective" in obj_type:
            obj_list.append(
                CalculableObjective(
                    name=input_template["Objectives"].loc["name"][obj_index],
                    obj_max_bool=input_template["Objectives"].loc["obj_max_bool"][obj_index],
                    lower_bound=input_template["Objectives"].loc["lower_bound"][obj_index],
                    upper_bound=input_template["Objectives"].loc["upper_bound"][obj_index],
                    transformer=input_template["Variables"].loc["transformer"][obj_index]
                    if not pd.isnull(input_template["Variables"].loc["transformer"][obj_index])
                    else None,
                    units=input_template["Objectives"].loc["units"][obj_index]
                    if not pd.isnull(input_template["Objectives"].loc["units"][obj_index])
                    else None,
                    obj_function=None,
                )
            )

    X = input_template["Variables"].iloc[5:, :].values
    Y = input_template["Objectives"].iloc[7:, :].values.astype(float)
    var_list = VariablesList(var_list)
    obj_list = ObjectivesList(obj_list)

    return X, Y, var_list, obj_list



import math
import os
import random
from attrs import define, field
from typing import List, Optional, Union

import numpy as np
import torch
from botorch.utils.sampling import get_polytope_samples
from scipy.stats import qmc

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.opt.constraints import ConstraintsList
from nemo_bo.opt.utils_samplers.sampling import *
from nemo_bo.opt.variables import ContinuousVariable, VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class SampleGenerator:
    """

    Base class for the sample generators

    Parameters
    ----------
    num_new_points: int = None
        The number of samples to generate

    """

    num_new_points: Optional[int] = None

    def generate_samples(self, *args, **kwargs) -> None:
        """

        Function that returns an X array of self.num_new_points number of samples

        """
        pass


@define
class PolytopeSampling(SampleGenerator):
    """

    Class that is used to generate samples from a polytope that is limited by the variables lower/upper bounds and
    constraints.  The sampler uses the implementation from the BoTorch package:
    M. Balandat, B. Karrer, D. R. Jiang, S. Daulton, B. Letham, A. G. Wilson, and E. Bakshy. BoTorch: A Framework for
    Efficient Monte-Carlo Bayesian Optimization. Advances in Neural Information Processing Systems 33, 2020.
    http://arxiv.org/abs/1910.06403
    https://github.com/pytorch/botorch
    https://botorch.org/

    """

    def generate_samples(
        self,
        variables: VariablesList,
        constraints: Optional[ConstraintsList] = None,
        seed: int = None,
        n_burnin: int = 10000,
        thinning: int = 32,
    ) -> np.ndarray:
        """

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        constraints: ConstraintsList, Default = None,
            ConstraintsList object that contains information about all constraints
        seed: int = None,
            Seed for a random number generator for the manual_seed function in botorch.utils.sampling
        n_burnin: int = 10000,
            The number of burn-in samples for the Markov chain sampler, where it drops the first n_burnin number of
            sampled points
        thinning: int = 32,
            The amount of thinning from the total number of generated by the hit-and-run Markov chain sampler

        Returns
        -------
        X: np.ndarray
            Array of generated samples that are within the defined bounds and passes all constraints

        """
        kwargs = {"seed": seed, "n_burnin": n_burnin, "thinning": thinning}
        if constraints is not None:
            for constr in constraints:
                if constr.constraint_type == "ineq":
                    kwargs["inequality_constraints"] = [constr.build_polytope_sampler_constraint()]
                else:
                    kwargs["equality_constraints"] = [constr.build_polytope_sampler_constraint()]

        X = (
            get_polytope_samples(
                n=self.num_new_points,
                bounds=torch.tensor([variables.lower_bounds, variables.upper_bounds]),
                **kwargs,
            )
            .cpu()
            .numpy()
        )

        if variables.num_cat_var > 0:
            return variables.descriptor_to_name(X)
        else:
            return X


@define
class LatinHyperCubeSampling(SampleGenerator):
    """

    Class that is used to generate samples via Latin Hypercube Sampling for continuous variables or a mixture of
    continuous and categorical variables. Samples generated will be limited by the variables lower/upper bounds,
    categorical variable levels, and constraints.  This sampler uses the implementation from the smt package by M. A.
    Bouhlel and J. T. Hwang and N. Bartoli and R. Lafage and J. Morlier and J. R. R. A. Martins, A Python surrogate
    modeling framework with derivatives. Advances in Engineering Software, 2019, 102662
    https://doi.org/10.1016/j.advengsoft.2019.03.005
    https://smt.readthedocs.io/en/latest/

    """

    def generate_samples(
        self, variables: VariablesList, constraints: Optional[ConstraintsList] = None, **kwargs
    ) -> np.ndarray:
        """

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        constraints: ConstraintsList, Default = None,
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        X: np.ndarray
            Array of generated samples that are within the defined bounds and passes all constraints

        """
        xlimits = np.array(
            [
                [var.lower_bound, var.upper_bound] if isinstance(var, ContinuousVariable) else var.categorical_levels
                for var in variables.variables
            ],
            dtype=object,
        )

        if variables.num_cat_var > 0:
            xtypes = [
                FLOAT if isinstance(var, ContinuousVariable) else (ENUM, var.num_categorical_levels)
                for var in variables.variables
            ]
            sampler = MixedIntegerSamplingMethod(xtypes, xlimits, LHS, random_state=random.randint(0, 1000000))

            X = sampler(self.num_new_points)

            X = X.astype(np.object)
            for var_index, var in enumerate(variables.variables):
                if not isinstance(var, ContinuousVariable):
                    X[:, var_index] = var.enum_to_cat_level(X[:, var_index])

            # To do: Improve the implementation for constraints for LHS.
            # Currently this does not sample the available space in a guaranteed LHS manner
            if constraints is not None:
                num_points_temp = self.num_new_points
                X = X[constraints.bool_evaluate_constraints(X)]
                while X.shape[0] < self.num_new_points:
                    num_points_temp = math.ceil(num_points_temp * 1.2)
                    X = sampler(num_points_temp)

                    X = X.astype(np.object)
                    for var_index, var in enumerate(variables.variables):
                        if not isinstance(var, ContinuousVariable):
                            X[:, var_index] = var.enum_to_cat_level(X[:, var_index])

                    X = X[constraints.bool_evaluate_constraints(X)]

            logger.info(f"Generated {X.shape[0]} mixed integer Latin hypercube sampling points in the variable space")

        else:
            sampler = LHS(xlimits=xlimits, **kwargs)

            X = sampler(self.num_new_points)

            # To do: Improve the implementation for constraints for LHS.
            # Currently this does not sample the available space in a guaranteed LHS manner after constraints are
            # applied
            if constraints is not None:
                num_points_temp = self.num_new_points
                X = X[constraints.bool_evaluate_constraints(X)]
                while X.shape[0] < self.num_new_points:
                    num_points_temp = math.ceil(num_points_temp * 1.2)
                    X = sampler(num_points_temp)
                    X = X[constraints.bool_evaluate_constraints(X)]

            logger.info(f"Generated {X.shape[0]} continuous Latin hypercube sampling points in the variable space")

        return X


@define
class RandomSampling(SampleGenerator):
    """

    Class that is used to generate samples via Random Sampling for continuous variables or a mixture of
    continuous and categorical variables. Samples generated will be limited by the variables lower/upper bounds,
    categorical variable levels, and constraints.  This sampler uses the implementation from the smt package by M. A.
    Bouhlel and J. T. Hwang and N. Bartoli and R. Lafage and J. Morlier and J. R. R. A. Martins, A Python surrogate
    modeling framework with derivatives. Advances in Engineering Software, 2019, 102662
    https://doi.org/10.1016/j.advengsoft.2019.03.005
    https://smt.readthedocs.io/en/latest/

    """

    def generate_samples(
        self, variables: VariablesList, constraints: Optional[ConstraintsList] = None, **kwargs
    ) -> np.ndarray:
        """

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        constraints: ConstraintsList, Default = None,
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        X: np.ndarray
            Array of generated samples that are within the defined bounds and passes all constraints

        """
        xlimits = np.array(
            [
                [var.lower_bound, var.upper_bound] if isinstance(var, ContinuousVariable) else var.categorical_levels
                for var in variables.variables
            ],
            dtype=object,
        )

        if variables.num_cat_var > 0:
            logger.info(f"Generating {self.num_new_points} mixed integer random sampling points in the variable space")
            xtypes = [
                FLOAT if isinstance(var, ContinuousVariable) else (ENUM, var.num_categorical_levels)
                for var in variables.variables
            ]
            sampler = MixedIntegerSamplingMethod(xtypes, xlimits, Random)
            X = sampler(self.num_new_points)

            X = X.astype(np.object)
            for var_index, var in enumerate(variables.variables):
                if not isinstance(var, ContinuousVariable):
                    X[:, var_index] = var.enum_to_cat_level(X[:, var_index])

            if constraints is not None:
                num_points_temp = self.num_new_points
                X = X[constraints.bool_evaluate_constraints(X)]
                while X.shape[0] < self.num_new_points:
                    num_points_temp = math.ceil(num_points_temp * 1.2)
                    X = sampler(num_points_temp)

                    X = X.astype(np.object)
                    for var_index, var in enumerate(variables.variables):
                        if not isinstance(var, ContinuousVariable):
                            X[:, var_index] = var.enum_to_cat_level(X[:, var_index])

                    X = X[constraints.bool_evaluate_constraints(X)]

        else:
            logger.info(f"Generating {self.num_new_points} continuous random sampling points in the variable space")
            sampler = Random(xlimits=xlimits, **kwargs)
            X = sampler(self.num_new_points)

            if constraints is not None:
                num_points_temp = self.num_new_points
                X = X[constraints.bool_evaluate_constraints(X)]
                while X.shape[0] < self.num_new_points:
                    num_points_temp = math.ceil(num_points_temp * 1.2)
                    X = sampler(num_points_temp)
                    X = X[constraints.bool_evaluate_constraints(X)]

        return X


@define
class SobolSampling(SampleGenerator):
    """

    Class that is used to generate scrambled Sobol sequences for continuous variables. Generator descriptor values will
    be converted to their corresponding categorical variables. Samples generated will be limited by the variables
    lower/upper bounds and constraints

    """

    def generate_samples(self, variables: VariablesList, constraints: Optional[ConstraintsList] = None) -> np.ndarray:
        """

        Parameters
        ----------
        variables: VariablesList
            VariablesList object that contains information about all variables
        constraints: ConstraintsList, Default = None,
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        X: np.ndarray
            Array of generated samples that are within the defined bounds and passes all constraints

        """
        sampler = qmc.Sobol(d=variables.n_var)

        num_samples_power = math.ceil(math.log2(self.num_new_points))
        if 2**num_samples_power != self.num_new_points:
            logger.info(
                f"Generating {2 ** num_samples_power} sobol sampling points in the variable "
                f"space (different to the desired {self.num_new_points} to guarantee the balance properties "
                f"of the sobol sequence)."
            )

        X = sampler.random_base2(m=num_samples_power)
        X = qmc.scale(X, variables.lower_bounds, variables.upper_bounds)

        # To do: Improve the implementation for constraints for Sobol sampling.
        # Currently this does not guarantee the balance properties of the sobol sequence after constraints are applied
        if constraints is not None:
            X = X[constraints.bool_evaluate_constraints(X)]
            while X.shape[0] < self.num_new_points:
                num_samples_power += 1
                sampler.reset()
                X = sampler.random_base2(m=num_samples_power)
                X = qmc.scale(X, variables.lower_bounds, variables.upper_bounds)
                X = X[constraints.bool_evaluate_constraints(X)]

            logger.info(f"Retained {X.shape[0]} sobol sampling points in the variable space after applying constraints")

        if variables.num_cat_var > 0:
            return variables.descriptor_to_name(X)
        else:
            return X


@define
class PoolBased:
    """

    Class that is used to store and take samples from designated X and Y arrays (a pool of samples). For simulated
    closed-loop optimisation problems, this is typically used to obtain true results whereas the alternative benchmark
    function would return predicted results instead

    Parameters
    ----------
    X_pool: np.ndarray
        The X array to be used in the sample pool
    Y_pool: np.ndarray
        The Y array to be used in the sample pool

    """

    X_pool: np.ndarray
    Y_pool: np.ndarray
    index: str = field(init=False)

    def __attrs_post_init__(self):
        self.index = None

    def generate_samples(self) -> np.ndarray:
        """

        Returns
        -------
        np.ndarray
            Array of X-values that are in the user-supplied pool

        """

        return self.X_pool

    def Y_and_update_pool(self, index: Union[int, List[int]]) -> np.ndarray:
        """

        Uses the index to select Y-values, delete them from the Y sample pool and return them

        Parameters
        ----------
        index: int, List[int]
            Index/indexes used to reference the Y sample pool

        Returns
        -------
        Y: np.ndarray
            The selected Y-values based on the index/indexes

        """
        if isinstance(index, int):
            index = [index]

        # Identifies the Y values to be returned
        Y = np.array([self.Y_pool[i] for i in index])

        # Removes the samples in the sample pool based on the supplied index/indexes
        self.X_pool = np.delete(self.X_pool, index, 0)
        self.Y_pool = np.delete(self.Y_pool, index, 0)

        return Y



import copy
import os
from attrs import define, field

from typing import List, Optional, Union

import numpy as np
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder, StandardScaler

import nemo_bo.utils.logger as logging_nemo

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class ContinuousVariable:
    """

    Class for a variable with continuous values

    Parameters
    ----------
    name: str
        Name of the variable
    lower_bound: int | float
        The lower bound of the variable
    upper_bound: int | float
        The upper bound of the variable
    transformer: Optional. Default = None
        The transformation that will be applied to the variable for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.
    units: str, Default = ""
        The units for the variable

    """

    name: str
    lower_bound: Union[int, float]
    upper_bound: Union[int, float]
    transformer: Optional = None
    units: str = ""

    def transform(self, X: np.ndarray) -> np.ndarray:
        """

        Fits a transformation scalar to the inputted X array when self.transformer is "normalisation",
        "standardisation" or a transformer object. The X array is subsequently transformed according to the
        fitted scalar. When self.transformation_type = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        elif self.transformer == "normalisation":
            self.transformer = MinMaxScaler()
        elif self.transformer == "standardisation":
            self.transformer = StandardScaler()

        return self.transformer.fit_transform(X)

    def transform_only(self, X: np.ndarray) -> np.ndarray:
        """

        The X array is transformed according to the fit scalar in the self.transformer object. When
        self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        else:
            return self.transformer.transform(X)

    def inverse_transform(self, X_transform: np.ndarray) -> np.ndarray:
        """

        The X_transform array undergoes an inverse transform according to the fit scalar in the self.transformer
        object. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X_transform: np.ndarray
            X array containing transformed values for one variable

        """

        if X_transform.ndim == 1:
            X_transform = X_transform.reshape(-1, 1)

        if self.transformer == "none":
            return X_transform
        else:
            return self.transformer.inverse_transform(X_transform)


@define(kw_only=True)
class CategoricalVariable:
    """

    Base class for categorical variables

    Parameters
    ----------
    name: str
        Name of the variable
    categorical_levels: List[str | int | float]
        The categories/choices/levels for a given categorical variable
    units: str, Default = ""
        The units for the variable

    """

    name: str
    categorical_levels: List[Union[str, int, float]]
    units: str = ""

    def enum_to_cat_level(self, X_array: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level index to its respective name/value

        Parameters
        ----------
        X_array: np.ndarray
            Array of indexes that correspond to the categorical levels for a categorical variable

        Returns
        -------
        np.ndarray
            Array where the indexes that correspond to the categorical levels have been replaced by the name/value

        """
        return np.array([self.categorical_levels[int(x)] for x in X_array])


@define
class CategoricalVariableOneHot(CategoricalVariable):
    """

    Class for categorical variables that will be treated using one-hot encoding

    Parameters
    ----------
    name: str
        Name of the variable
    categorical_levels: List[str | int | float]
        The categories/choices/levels for a given categorical variable
    units: str, Default = ""
        The units for the variable

    """

    num_categorical_levels: int = field(init=False)

    def __attrs_post_init__(self):
        self.num_categorical_levels = len(self.categorical_levels)

    def one_hot_encode(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical levels in an X array to a one-hot-encoding array

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical levels as names/values

        Returns
        -------
        np.ndarray
            Array where the categorical levels have been converted using one-hot encoding

        """
        return OneHotEncoder(handle_unknown="ignore", categories=self.categorical_levels).transform(X).toarray()


@define(kw_only=True)
class CategoricalVariableDiscreteValues(CategoricalVariable):
    """

    Class for categorical variables that have levels of discrete values

    Parameters
    ----------
    name: str
        Name of the variable
    categorical_levels: List[str | int | float]
        The categories/choices/levels for a given categorical variable
    units: str, Default = ""
        The units for the variable
    lower_bound: int | float, Default = None
        The lower bound of the variable
    upper_bound: int | float, Default = None
        The upper bound of the variable
    transformer: Optional. Default = None
        The transformation that should be applied to the variable for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.

    """

    lower_bound: Optional[Union[int, float]] = None
    upper_bound: Optional[Union[int, float]] = None
    transformer: Optional = None
    num_categorical_levels: int = field(init=False)

    def __attrs_post_init__(self):
        self.num_categorical_levels = len(self.categorical_levels)
        if self.lower_bound is None:
            self.lower_bound = min(self.categorical_levels)

        if self.upper_bound is None:
            self.upper_bound = max(self.categorical_levels)

    def transform(self, X: np.ndarray) -> np.ndarray:
        """

        Fits a transformation scalar to the inputted X array when self.transformer is "normalisation",
        "standardisation" or a transformer object. The X array is subsequently transformed according to the
        fitted scalar. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        elif self.transformer == "normalisation":
            self.transformer = MinMaxScaler()
        elif self.transformer == "standardisation":
            self.transformer = StandardScaler()

        return self.transformer.fit_transform(X)

    def transform_only(self, X: np.ndarray) -> np.ndarray:
        """

        The X array is transformed according to the fit scalar in the self.transformer object. When
        self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        else:
            return self.transformer.transform(X)

    def inverse_transform(self, X_transform: np.ndarray) -> np.ndarray:
        """

        The X_transform array undergoes an inverse transform according to the fit scalar in the self.transformer
        object. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X_transform: np.ndarray
            X array containing transformed values for one variable

        """
        if X_transform.ndim == 1:
            X_transform = X_transform.reshape(-1, 1)

        if self.transformer == "none":
            return X_transform
        else:
            return self.transformer.inverse_transform(X_transform)

    def value_to_discrete(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level index to their closest discrete value

        Parameters
        ----------
        X_array: np.ndarray
            Array containing X values to be converted

        Returns
        -------
        np.ndarray
            Array where the X values have been converted to their closest discrete values

        """

        X_new = copy.copy(X)
        for index, value in enumerate(X):
            diff = ((np.array(self.categorical_levels) - value) ** 2) ** 0.5
            diff_min_index = np.argmin(diff)
            X_new[index] = self.categorical_levels[diff_min_index]

        return X_new

    def enum_to_discrete(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level indexes to the corresponding discrete values

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level indexes to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level indexes have been converted to their corresponding discrete values

        """
        return np.vstack([self.categorical_descriptors[int(x)] for x in X])


@define(kw_only=True)
class CategoricalVariableWithDescriptors(CategoricalVariable):
    """

    Class for categorical variables that have levels which are arrays of descriptor values

    Parameters
    ----------
    name: str
        Name of the variable
    categorical_levels: List[str | int | float]
        The categories/choices/levels for a given categorical variable
    units: str, Default = ""
        The units for the variable
    descriptor_names: List[str]
        The names for each type of descriptor in the descriptor arrays
    categorical_descriptors: List[Any] | np.ndarray
        The array of descriptor values where the shape is [{number of categorical levels}, {number of descriptor types}],
        where each row in the array are the descriptors for each categorical level
    transformer: Optional. Default = None
        The transformation that should be applied to the variable for modelling. When left as the default
        None, the modelling uses the default transformations specified by the model type. Can be specified as the
        strings "normalisation", "standardisation", or "none" for min-max normalisation, standardisation to a mean of 0
        and a standard deviation of 1, or no transformation respectively. Alternatively, a class instance with
        fit_transform, transform, and inverse_transform functions, such as a scikit-learn transformer can be provided
        directly too.
    lower_bound: int | float, Default = None
        The lower bound of the variable
    upper_bound: int | float, Default = None
        The upper bound of the variable

    """

    descriptor_names: List[str]
    categorical_descriptors: Union[List[List[Union[int, float]]], np.ndarray]
    transformer: Optional = None
    lower_bound: Optional[Union[int, float]] = None
    upper_bound: Optional[Union[int, float]] = None
    num_categorical_levels: int = field(init=False)
    num_descriptors: int = field(init=False)

    def __attrs_post_init__(self):
        self.num_categorical_levels = len(self.categorical_levels)

        if not isinstance(self.categorical_descriptors, np.ndarray):
            self.categorical_descriptors = np.array(self.categorical_descriptors)

        if self.categorical_descriptors.ndim != 2:
            raise ValueError(
                f"The categorical descriptor array was provided is a {self.categorical_descriptors.ndim}D "
                f"array. Please ensure that it is a 2D array"
            )

        if self.lower_bound is None:
            self.lower_bound = []
            descriptor_range = np.ptp(self.categorical_descriptors, axis=0)
            for descriptor, range in zip(self.categorical_descriptors.T, descriptor_range):
                self.lower_bound.append(descriptor.min() - (range * 0.10))

        if self.upper_bound is None:
            self.upper_bound = []
            descriptor_range = np.ptp(self.categorical_descriptors, axis=0)
            for descriptor, range in zip(self.categorical_descriptors.T, descriptor_range):
                self.upper_bound.append(descriptor.max() + (range * 0.10))

        self.num_descriptors = self.categorical_descriptors.shape[1]

        if len(self.descriptor_names) != self.num_descriptors:
            raise ValueError(
                f"The number of names provided for the descriptors ({len(self.descriptor_names)}) is not "
                f"the same as the number of descriptors types provided ({self.num_descriptors})"
            )

    def name_to_descriptor(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level names into an array of the corresponding descriptor values

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level names to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level names have been converted to their corresponding descriptor values

        """
        return self.categorical_descriptors[self.categorical_levels.index(X)]

    def name_to_enum(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level names to the corresponding categorical level index

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level names to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level names have been converted to their corresponding categorical level index

        """
        return np.vstack([self.categorical_levels.index(x) for x in X])

    def enum_to_descriptor(self, X: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level indexes to the corresponding descriptor values

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level indexes to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level indexes have been converted to their corresponding descriptor values

        """
        return np.vstack([self.categorical_descriptors[int(x)] for x in X])

    def transform(self, X: np.ndarray) -> np.ndarray:
        """

        Fits a transformation scalar to the inputted X array when self.transformer is "normalisation",
        "standardisation" or a transformer object. The X array is subsequently transformed according to the
        fitted scalar. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        elif self.transformer == "normalisation":
            self.transformer = MinMaxScaler()
        elif self.transformer == "standardisation":
            self.transformer = StandardScaler()

        return self.transformer.fit_transform(X)

    def transform_only(self, X: np.ndarray) -> np.ndarray:
        """

        The X array is transformed according to the fit scalar in the self.transformer object. When
        self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for one variable

        """
        if self.transformer is None:
            raise AttributeError(
                "Please either: 1. Fit the model first, 2. Set transformer = 'none', 'normalisation', or "
                "'standarisation' when creating the variable instance, or 3. Pass a transformer such as a scikit-learn "
                "transformer when creating the variable instance"
            )

        if X.ndim == 1:
            X = X.reshape(-1, 1)

        if self.transformer == "none":
            return X
        else:
            return self.transformer.transform(X)

    def inverse_transform(self, X_transform: np.ndarray) -> np.ndarray:
        """

        The X_transform array undergoes an inverse transform according to the fit scalar in the self.transformer
        object. When self.transformer = "none", no transformation occurs

        Parameters
        ----------
        X_transform: np.ndarray
            X array containing transformed values for one variable

        """

        if X_transform.ndim == 1:
            X_transform = X_transform.reshape(-1, 1)

        if self.transformer == "none":
            return X_transform
        else:
            return self.transformer.inverse_transform(X_transform)


@define
class VariablesList:
    """

    Class that stores all variables and their respective properties

    Parameters
    ----------
    objectives: List[Objective]
        A list of ContinuousVariable and/or CategoricalVariable objects

    """

    variables: List[Union[ContinuousVariable, CategoricalVariable]]
    names: List[str] = field(init=False)
    units: List[str] = field(init=False)
    num_cat_onehot_var: int = field(init=False)
    num_cat_discrete_var: int = field(init=False)
    num_cat_descriptor_var: int = field(init=False)
    num_cat_var: int = field(init=False)
    categorical_levels: List[List[Union[str, int]]] = field(init=False)
    num_categorical_levels: int = field(init=False)
    descriptor_names: List[List[str]] = field(init=False)
    categorical_descriptors: List[List[str]] = field(init=False)
    num_descriptors: List[int] = field(init=False)
    cont_var_units: List[str] = field(init=False)
    lower_bounds: List[Union[int, float]] = field(init=False)
    upper_bounds: List[Union[int, float]] = field(init=False)
    var_splitter_indexes: List[int] = field(init=False)
    n_var: int = field(init=False)
    continuous_var_names: List[str] = field(init=False)

    def __attrs_post_init__(self):
        self.names = [var.name for var in self.variables]
        self.units = [var.units for var in self.variables]

        # Counts the number of each type of categorical variables, and the total number of categorical variables
        self.num_cat_onehot_var = sum(isinstance(var, CategoricalVariableOneHot) for var in self.variables)
        self.num_cat_discrete_var = sum(isinstance(var, CategoricalVariableDiscreteValues) for var in self.variables)
        self.num_cat_descriptor_var = sum(isinstance(var, CategoricalVariableWithDescriptors) for var in self.variables)
        self.num_cat_var = self.num_cat_onehot_var + self.num_cat_discrete_var + self.num_cat_descriptor_var

        # List of the categorical levels for every categorical variables
        self.categorical_levels = [
            var.categorical_levels if not isinstance(var, ContinuousVariable) else None for var in self.variables
        ]
        # List of the number of categorical levels for every categorical variables
        self.num_categorical_levels = [
            var.num_categorical_levels if not isinstance(var, ContinuousVariable) else 0 for var in self.variables
        ]
        # List of descriptor names for every categorical variable with descriptors
        self.descriptor_names = [
            var.descriptor_names if isinstance(var, CategoricalVariableWithDescriptors) else None
            for var in self.variables
        ]
        # List of descriptor value arrays for every categorical variable with descriptors
        self.categorical_descriptors = [
            var.categorical_descriptors if isinstance(var, CategoricalVariableWithDescriptors) else None
            for var in self.variables
        ]
        # List of the number of descriptors in every categorical variable with descriptors
        self.num_descriptors = [
            var.num_descriptors if isinstance(var, CategoricalVariableWithDescriptors) else 0 for var in self.variables
        ]

        # Creates lists of lower and upper bounds for all variables (where applicable), and a list that contains
        # indexes that indicate where to split the X array to separate each variable
        (self.lower_bounds, self.upper_bounds, self.var_splitter_indexes, categorical_descriptor_index_range,) = (
            [],
            [],
            [],
            [],
        )
        var_counter = 0
        self.cont_var_units = []
        for var_index, var in enumerate(self.variables):
            if isinstance(var, ContinuousVariable) or isinstance(var, CategoricalVariableDiscreteValues):
                self.lower_bounds.append(var.lower_bound)
                self.upper_bounds.append(var.upper_bound)
                self.cont_var_units.append(var.units)
                var_counter += 1
            elif isinstance(var, CategoricalVariableOneHot):
                var_counter += var.num_categorical_levels
            elif isinstance(var, CategoricalVariableWithDescriptors):
                for lower_value, upper_value in zip(var.lower_bound, var.upper_bound):
                    self.lower_bounds.append(lower_value)
                    self.upper_bounds.append(upper_value)

                categorical_descriptor_index_range.append(
                    [
                        var.name,
                        var_index,
                        var_counter,
                        var_counter + var.num_descriptors,
                    ]
                )
                var_counter += var.num_descriptors
                self.cont_var_units += [""] * var.num_descriptors

            if var_index + 1 < len(self.variables):
                self.var_splitter_indexes.append(var_counter)

        self.n_var = var_counter

        # List of variables names for all continuous variables, categorical variables with discrete values, and the
        # descriptor names in categorical variables with descriptors
        self.continuous_var_names = []
        for var in self.variables:
            if isinstance(var, ContinuousVariable) or isinstance(var, CategoricalVariableDiscreteValues):
                self.continuous_var_names.append(var.name)
            elif isinstance(var, CategoricalVariableWithDescriptors):
                for descriptor_name in var.descriptor_names:
                    self.continuous_var_names.append(descriptor_name)

    def transform(self, X: np.ndarray, transform_type: Optional = None) -> np.ndarray:
        """

        Fits and transforms the X array containing data for multiple variables based on each variables'
        transformer attribute

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all variables
        transform_type: str, Default = None
            Assigns the transformation type if one was not specified when the variable object was created

        Returns
        -------
        np.ndarray
            X array containing transformed values for all variables

        """
        X_split = self.split_var(X)
        for var_index, var in enumerate(self.variables):
            if not isinstance(var, CategoricalVariableOneHot):
                if var.transformer is None:
                    var.transformer = transform_type

                if X_split[var_index].shape[1] == 1:
                    X_split[var_index] = var.transform(X_split[var_index]).reshape(-1, 1)
                else:
                    X_split[var_index] = var.transform(X_split[var_index])

        return np.hstack(X_split)

    def transform_only(self, X: np.ndarray) -> np.ndarray:
        """

        Transforms the X array containing data for multiple variables based on each variables' transformer
        attribute using the previously fitted scalars

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all variables

        Returns
        -------
        np.ndarray
            X array containing transformed values for all variables

        """
        X_split = self.split_var(X)
        for var_index, var in enumerate(self.variables):
            if not isinstance(var, CategoricalVariableOneHot):
                if X_split[var_index].shape[1] == 1:
                    X_split[var_index] = var.transform_only(X_split[var_index]).reshape(-1, 1)
                else:
                    X_split[var_index] = var.transform_only(X_split[var_index])

        return np.hstack(X_split)

    def inverse_transform(self, X: np.ndarray) -> np.ndarray:
        """

        The inputted X array containing data for multiple variables undergoes an inverse transformaion using the
        previously fitted scalars

        Parameters
        ----------
        X: np.ndarray
            X array containing transformed values for all variables

        Returns
        -------
        np.ndarray
            X array containing untransformed values for all variables

        """
        X_split = self.split_var(X)
        for var_index, var in enumerate(self.variables):
            if not isinstance(var, CategoricalVariableOneHot):
                if X_split[var_index].shape[1] == 1:
                    X_split[var_index] = var.inverse_transform(X_split[var_index]).reshape(-1, 1)
                else:
                    X_split[var_index] = var.inverse_transform(X_split[var_index])

        return np.hstack(X_split)

    def categorical_transform(self, X: np.ndarray) -> np.ndarray:
        """

        Transforms all categorical data in the X array according to their categorical variable type:
        - CategoricalVariableOneHot: Converts the categorical levels in an X array to a one-hot-encoding array
        - CategoricalVariableDiscreteValues: Converts the categorical level index to their closest discrete value
        - CategoricalVariableWithDescriptors: Converts the categorical level names into an array of the corresponding
        descriptor values

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        Returns
        -------
        np.ndarray
            X array containing transformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        """
        X_transform = np.zeros_like(X, dtype=np.object)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableOneHot):
                for index in range(X_transform[:, var_index].shape[0]):
                    X_transform[index, var_index] = var.one_hot_encode(X[index, var_index])
            elif isinstance(var, CategoricalVariableDiscreteValues):
                X_transform[:, var_index] = var.value_to_discrete(X[:, var_index])
            elif isinstance(var, CategoricalVariableWithDescriptors):
                for index in range(X_transform[:, var_index].shape[0]):
                    X_transform[index, var_index] = var.name_to_descriptor(X[index, var_index])
            elif isinstance(var, ContinuousVariable):
                X_transform[:, var_index] = X[:, var_index]

        return np.array([np.hstack(row) for row in X_transform]).astype(float)

    def descriptor_to_name(self, X: np.ndarray) -> np.ndarray:
        """

        Transforms all X values that correspond to descriptor values for categorical variables to their respective
        categorical level name according to the shortest normalised euclidean distance

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        Returns
        -------
        np.ndarray
            X array containing transformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        """
        X_split = self.split_var(X)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableWithDescriptors):
                cat_name_list = []
                for x in X_split[var_index]:
                    # Normalises the x array
                    x_norm = (x - np.array(var.lower_bound)) / (np.array(var.upper_bound) - np.array(var.lower_bound))
                    # Normalises this categorical variable's descriptor array
                    norm_categorical_descriptors = (var.categorical_descriptors - np.array(var.lower_bound)) / (
                        np.array(var.upper_bound) - np.array(var.lower_bound)
                    )
                    # Calculates the normalised euclidean distance for the normalised x array against every categorical
                    # level's normalised descriptor array
                    euc_dist = [np.linalg.norm(cat_level - x_norm) for cat_level in norm_categorical_descriptors]
                    # Identifies the categorical level index with the shortest normalised euclidean distance
                    euc_dist_min_index = np.argmin(euc_dist)
                    cat_name_list.append([var.categorical_levels[euc_dist_min_index]])
                X_split[var_index] = np.array(cat_name_list)

        return np.hstack(X_split)

    def categorical_values_euc(self, X: np.ndarray) -> np.ndarray:
        """

        Transforms all X values that correspond to descriptor
         values or discrete values for categorical variables to
        their categorical value(s) according to the shortest normalised euclidean distance

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        Returns
        -------
        np.ndarray
            X array containing transformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        """
        X_split = self.split_var(X)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableDiscreteValues):
                X_split[var_index] = var.value_to_discrete(X_split[var_index])
        X = np.hstack(X_split).astype(np.float)

        if any(isinstance(var, CategoricalVariableWithDescriptors) for var in self.variables):
            X = self.value_to_descriptor(X)

        return X

    def value_to_descriptor(self, X_array: np.ndarray) -> np.ndarray:
        """

        Transforms all X values that correspond to descriptors for categorical variables to descriptors values that are
        at the shortest normalised euclidean distance

        Parameters
        ----------
        X: np.ndarray
            X array containing untransformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        Returns
        -------
        np.ndarray
            X array containing transformed values for all categorical variables. Any continuous variable data in the
            array are unaffected

        """
        X_split = self.split_var(X_array)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableWithDescriptors):
                cat_descriptor_list = []
                for x in X_split[var_index]:
                    x_norm = (x - np.array(var.lower_bound)) / (np.array(var.upper_bound) - np.array(var.lower_bound))
                    norm_categorical_descriptors = (var.categorical_descriptors - np.array(var.lower_bound)) / (
                        np.array(var.upper_bound) - np.array(var.lower_bound)
                    )
                    euc_dist = [np.linalg.norm(cat_level - x_norm) for cat_level in norm_categorical_descriptors]
                    euc_dist_min_index = np.argmin(euc_dist)
                    cat_descriptor_list.append(var.categorical_descriptors[euc_dist_min_index])
                X_split[var_index] = np.array(cat_descriptor_list)

        return np.hstack(X_split)

    def enum_to_values(self, X_array: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level indexes in the X values for categorical variables with discrete or descriptor
        values to the corresponding discrete or descriptor values

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level indexes to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level indexes have been converted to their corresponding descriptor values

        """
        X_array_object = X_array.astype(np.object)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableWithDescriptors):
                X_array_object[:, var_index] = var.enum_to_descriptor(X_array[:, var_index])
            elif isinstance(var, CategoricalVariableDiscreteValues):
                X_array_object[:, var_index] = var.enum_to_discrete(X_array[:, var_index])

        return X_array_object.astype("float64")

    def name_to_descriptor(self, X_array: np.ndarray) -> np.ndarray:
        """

        Converts the categorical level names in the X values for categorical variables with descriptor values to the
        corresponding descriptor arrays

        Parameters
        ----------
        X_array: np.ndarray
            Array containing categorical level indexes to be converted

        Returns
        -------
        np.ndarray
            Array where the categorical level indexes have been converted to their corresponding descriptor arrays

        """
        X_array_object = X_array.astype(np.object)
        for var_index, var in enumerate(self.variables):
            if isinstance(var, CategoricalVariableWithDescriptors):
                X_array_object[:, var_index] = var.name_to_descriptor(X_array[:, var_index])

        return X_array_object.astype("float64")

    def split_var(self, X_array: np.ndarray) -> List[np.ndarray]:
        """

        Splits an X-array that contains the continuous and categorical variables into a list of sub-arrays for each
        individual variable

        Parameters
        ----------
        X_array: np.ndarray
             X-array that contains the continuous and categorical variables, where the categorical variables are
             specified in their categorical levels (and not as descriptors for example)

        Returns
        -------
        List[np.ndarray]
            List of sub-arrays for each individual variable

        """
        return np.split(X_array, self.var_splitter_indexes, axis=1)

    def append(self, var: Union[ContinuousVariable, CategoricalVariable]):
        """

        Function for appending an additional variable to the current VariablesList object

        Parameters
        ----------
        var: ContinuousVariable | CategoricalVariable
            ContinuousVariable or CategoricalVariable variable

        """
        var_list = copy.copy(self.variables)
        var_list.append(var)

        return VariablesList(var_list)



# The NSGA-based methods were copied and adapted from J. Blank and K. Deb, pymoo: Multi-Objective Optimization in
# Python, in IEEE Access, vol. 8, pp. 89497-89509, 2020, doi: 10.1109/ACCESS.2020.2990567
# https://ieeexplore.ieee.org/document/9078759
# https://github.com/anyoptimization/pymoo

import numpy as np

from pymoo.core.survival import Survival
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.tournament import compare, TournamentSelection
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.util.dominator import Dominator
from pymoo.util.misc import find_duplicates, has_feasible
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.util.randomized_argsort import randomized_argsort

from nemo_bo.acquisition_functions.nsga_improvement.genetic import GeneticAlgorithm


# ---------------------------------------------------------------------------------------------------------
# Binary Tournament Selection Function
# ---------------------------------------------------------------------------------------------------------


def binary_tournament(pop, P, algorithm, **kwargs):
    n_tournaments, n_parents = P.shape

    if n_parents != 2:
        raise ValueError("Only implemented for binary tournament!")

    tournament_type = algorithm.tournament_type
    S = np.full(n_tournaments, np.nan)

    for i in range(n_tournaments):

        a, b = P[i, 0], P[i, 1]
        a_cv, a_f, b_cv, b_f = pop[a].CV[0], pop[a].F, pop[b].CV[0], pop[b].F
        rank_a, cd_a = pop[a].get("rank", "crowding")
        rank_b, cd_b = pop[b].get("rank", "crowding")

        # if at least one solution is infeasible
        if a_cv > 0.0 or b_cv > 0.0:
            S[i] = compare(a, a_cv, b, b_cv, method="smaller_is_better", return_random_if_equal=True)

        # both solutions are feasible
        else:

            if tournament_type == "comp_by_dom_and_crowding":
                rel = Dominator.get_relation(a_f, b_f)
                if rel == 1:
                    S[i] = a
                elif rel == -1:
                    S[i] = b

            elif tournament_type == "comp_by_rank_and_crowding":
                S[i] = compare(a, rank_a, b, rank_b, method="smaller_is_better")

            else:
                raise Exception("Unknown tournament type.")

            # if rank or domination relation didn't make a decision compare by crowding
            if np.isnan(S[i]):
                S[i] = compare(a, cd_a, b, cd_b, method="larger_is_better", return_random_if_equal=True)

    return S[:, None].astype(int, copy=False)


# ---------------------------------------------------------------------------------------------------------
# Survival Selection
# ---------------------------------------------------------------------------------------------------------


class RankAndCrowdingSurvival(Survival):
    def __init__(self, nds=None) -> None:
        super().__init__(filter_infeasible=True)
        self.nds = nds if nds is not None else NonDominatedSorting()

    def _do(self, problem, pop, *args, n_survive=None, **kwargs):

        # get the objective space values and objects
        F = pop.get("F").astype(float, copy=False)

        # the final indices of surviving individuals
        survivors = []

        # do the non-dominated sorting until splitting front
        fronts = self.nds.do(F, n_stop_if_ranked=n_survive)

        for k, front in enumerate(fronts):

            # calculate the crowding distance of the front
            crowding_of_front = calc_crowding_distance(F[front, :])

            # save rank and crowding in the individual class
            for j, i in enumerate(front):
                pop[i].set("rank", k)
                pop[i].set("crowding", crowding_of_front[j])

            # current front sorted by crowding distance if splitting
            if len(survivors) + len(front) > n_survive:
                I = randomized_argsort(crowding_of_front, order="descending", method="numpy")
                I = I[: (n_survive - len(survivors))]

            # otherwise take the whole front unsorted
            else:
                I = np.arange(len(front))

            # extend the survivors by all or selected individuals
            survivors.extend(front[I])

        return pop[survivors]


# =========================================================================================================
# Implementation
# =========================================================================================================


class NSGA2(GeneticAlgorithm):
    def __init__(
        self,
        pop_size=100,
        sampling=FloatRandomSampling(),
        selection=TournamentSelection(func_comp=binary_tournament),
        crossover=SBX(eta=15, prob=0.9),
        mutation=PM(eta=20),
        survival=RankAndCrowdingSurvival(),
        output=MultiObjectiveOutput(),
        **kwargs
    ):
        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            survival=survival,
            output=output,
            advance_after_initial_infill=True,
            **kwargs
        )

        self.termination = DefaultMultiObjectiveTermination()
        self.tournament_type = "comp_by_dom_and_crowding"

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]


def calc_crowding_distance(F, filter_out_duplicates=True):
    n_points, n_obj = F.shape

    if n_points <= 2:
        return np.full(n_points, np.inf)

    else:

        if filter_out_duplicates:
            # filter out solutions which are duplicates - duplicates get a zero finally
            is_unique = np.where(np.logical_not(find_duplicates(F, epsilon=1e-32)))[0]
        else:
            # set every point to be unique without checking it
            is_unique = np.arange(n_points)

        # index the unique points of the array
        _F = F[is_unique]

        # sort each column and get index
        I = np.argsort(_F, axis=0, kind="mergesort")

        # sort the objective space values for the whole matrix
        _F = _F[I, np.arange(n_obj)]

        # calculate the distance from each point to the last and next
        dist = np.row_stack([_F, np.full(n_obj, np.inf)]) - np.row_stack([np.full(n_obj, -np.inf), _F])

        # calculate the norm for each objective - set to NaN if all values are equal
        norm = np.max(_F, axis=0) - np.min(_F, axis=0)
        norm[norm == 0] = np.nan

        # prepare the distance to last and next vectors
        dist_to_last, dist_to_next = dist, np.copy(dist)
        dist_to_last, dist_to_next = dist_to_last[:-1] / norm, dist_to_next[1:] / norm

        # if we divide by zero because all values in one columns are equal replace by none
        dist_to_last[np.isnan(dist_to_last)] = 0.0
        dist_to_next[np.isnan(dist_to_next)] = 0.0

        # sum up the distance to next and last and norm by objectives - also reorder from sorted list
        J = np.argsort(I, axis=0)
        _cd = np.sum(dist_to_last[J, np.arange(n_obj)] + dist_to_next[J, np.arange(n_obj)], axis=1) / n_obj

        # save the final vector which sets the crowding distance for duplicates to zero to be eliminated
        crowding = np.zeros(n_points)
        crowding[is_unique] = _cd

    # crowding[np.isinf(crowding)] = 1e+14
    return crowding


parse_doc_string(NSGA2.__init__)



import copy
import os
import time
from dataclasses import dataclass
from itertools import combinations
from joblib import Parallel, delayed
from typing import Optional, Tuple, Union

import numpy as np
import random
import torch
from botorch.acquisition.multi_objective.monte_carlo import qNoisyExpectedHypervolumeImprovement
from botorch.acquisition.monte_carlo import qNoisyExpectedImprovement
from botorch.sampling.samplers import SobolQMCNormalSampler
from botorch.models.model_list_gp_regression import ModelListGP
from botorch.optim.optimize import optimize_acqf
from gpytorch.mlls.sum_marginal_log_likelihood import SumMarginalLogLikelihood
from scipy import spatial
from scipy.optimize import Bounds
from scipy.optimize import minimize as scipy_minimize
from scipy.stats import norm

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.acquisition_functions.base_acq_function import AcquisitionFunction
from nemo_bo.acquisition_functions.expected_improvement.ehvi import ExpectedHypervolumeImprovement
from nemo_bo.opt.constraints import ConstraintsList, LinearConstraint
from nemo_bo.opt.objectives import ObjectivesList, RegressionObjective
from nemo_bo.opt.samplers import PoolBased, SampleGenerator, PolytopeSampling
from nemo_bo.opt.variables import VariablesList
from nemo_bo.utils.data_proc import remove_all_nan_rows, remove_nan

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@dataclass
class ExpectedImprovement(AcquisitionFunction):
    """

    Class to instantiate to use the acquisition functions expected improvement for single-objective problems or
    expected hypervolume improvement for multi-objective problems

    Parameters
    ----------
    num_candidates: int
        The number of sets of X arrays to be suggested by the acquisition function
    num_restarts: int, Default = 4
        The number of times to restart the acquisition function. The result from every restart is saved, and the one
        with best improvement is selected
    exploration_factor: int, Default = None
        Determines the amount of exploration during optimisation by controlling the size of the effect of the standard
        deviation of a prediction on exploitation/exploration trade-off. Higher exploration_factor values lead to
        more exploration. Default for single objective optimisation is 1 and multi-objectivve optimisation is 3
    force_botorch_ei_methods: bool, Default = False
        Whether to force the expected improvement type methods to use the BoTorch-based ones, as long as the following
        conditions are met:
        1. All objectives are modelled with GPs
        2. All constraints, if used, are LinearConstraint objects
        3. The sampler chosen is not a PoolBased sampler

    """

    num_candidates: int
    num_restarts: Optional[int] = 4
    exploration_factor: Optional[int] = None
    force_botorch_ei_methods: bool = False

    def generate_candidates(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        sampler: Union[SampleGenerator, PoolBased],
        constraints: Optional[ConstraintsList] = None,
        **kwargs,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that is called to generate the best candidates using the expected improvement or expected
        hypervolume acquisition functions. This selection is made automatically based on the number of objectives

        Parameters
        ----------
        X: np.ndarray
            An array containing X-values for all variables
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        sampler: SampleGenerator | PoolBased
            Used during the ei or ehvi functions to generate samples (SampleGenerator) or provide the samples defined
            by the user (PoolBased)
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement or expected
            hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement or expected
            hypervolume improvement

        """
        self.sampler = sampler

        X, Y = remove_nan(X, Y)

        if constraints is not None:
            if all([isinstance(constr, LinearConstraint) for constr in constraints.constraints]):
                none_or_linear_constraints_bool = True
            else:
                none_or_linear_constraints_bool = False
        else:
            none_or_linear_constraints_bool = True

        if all([isinstance(obj, RegressionObjective) for obj in objectives.objectives]):
            all_gp_bool = all([obj == "gp" for obj in [obj.obj_function.name for obj in objectives.objectives]])
        else:
            all_gp_bool = False

        if self.force_botorch_ei_methods:
            if not all_gp_bool:
                logger.warning(
                    f"The model(s) for the optimisation are not all GP models therefore NEMO-based expected improvement"
                    f"methods will be used instead of the desired BoTorch methods"
                )
            if not none_or_linear_constraints_bool:
                logger.warning(
                    f"Only linear constraints are currently supported with the BoTorch methods within the NEMO "
                    f"framework therefore NEMO-based expected improvement methods will be used instead of the desired "
                    f"BoTorch methods"
                )
            if isinstance(self.sampler, PoolBased):
                logger.warning(
                    f"Pool-based samplers for benchmarking is not supported with the BoTorch methods within the NEMO "
                    f"framework therefore NEMO-based expected improvement methods will be used instead of the desired "
                    f"BoTorch methods"
                )

        if (
            self.force_botorch_ei_methods
            and all_gp_bool
            and none_or_linear_constraints_bool
            and not isinstance(self.sampler, PoolBased)
        ):
            # Use BoTorch based acquisition functions
            if objectives.n_obj > 1:
                selected_X, selected_Y = self.qnehvi_botorch(X, Y, variables, objectives, constraints)

            elif objectives.n_obj == 1:
                selected_X, selected_Y = self.qnei_botorch(X, Y, variables, objectives, constraints)

        else:
            # Use NEMO based acquisition functions
            if isinstance(self.sampler, PoolBased):
                if self.num_candidates > 4:
                    logger.warning(
                        f"The number of suggested candidates is 5 or greater, which is not recommended for pool-based "
                        f"samples with more than 60 samples due to long run times"
                    )

            if objectives.n_obj > 1:
                if isinstance(self.sampler, SampleGenerator):
                    if self.num_candidates == 1:
                        self.sampler.num_new_points = 262144
                    elif self.num_candidates == 2:
                        self.sampler.num_new_points = 512
                    elif self.num_candidates == 3:
                        self.sampler.num_new_points = 64
                    elif self.num_candidates == 4:
                        self.sampler.num_new_points = 32
                    elif self.num_candidates == 5:
                        self.sampler.num_new_points = 16
                    logger.info(
                        f"Setting the number of candidates to be generated by the sampler to "
                        f"{self.sampler.num_new_points} for more memory-efficient EHVI calculations"
                    )

                if self.exploration_factor is None:
                    self.exploration_factor = 3

                selected_X, selected_Y = self.ehvi(Y, variables, objectives, constraints)

            elif objectives.n_obj == 1:
                if isinstance(self.sampler, SampleGenerator):
                    if self.sampler.num_new_points is None:
                        if isinstance(self.sampler, PolytopeSampling):
                            self.sampler.num_new_points = 1024
                        else:
                            self.sampler.num_new_points = 262144
                    logger.info(
                        f"Setting the number of candidates to be generated by the sampler to "
                        f"{self.sampler.num_new_points} for more efficient EI calculations"
                    )

                if self.exploration_factor is None:
                    self.exploration_factor = 1

                selected_X, selected_Y = self.ei(Y, variables, objectives, constraints)

        return selected_X, selected_Y

    def ei(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected improvement

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement

        """
        logger.info(f"Identifying conditions with the highest expected improvement")

        success_res_fun_list = []
        success_selected_X_list = []
        success_selected_Y_list = []
        fail_res_fun_list = []
        fail_selected_X_list = []
        fail_selected_Y_list = []
        for iteration in range(self.num_restarts):
            start_time = time.perf_counter()

            if isinstance(self.sampler, PoolBased):
                logger.info(f"Returning X values from the provided sample pool")
                X_new_points = self.sampler.X_pool
            elif isinstance(self.sampler, SampleGenerator):
                X_new_points = self.sampler.generate_samples(variables, constraints)

            Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

            Y_norm = (Y - objectives.bounds[0, 0]) / (objectives.bounds[0, 1] - objectives.bounds[0, 0])
            Y_new_points_norm = (Y_new_points - objectives.bounds[0, 0]) / (
                objectives.bounds[0, 1] - objectives.bounds[0, 0]
            )
            # Min_max normalisation of Y standard deviation
            # I think this should be fine to minmax normalise like this
            Y_new_points_stddev_norm = Y_new_points_stddev / (objectives.bounds[0, 1] - objectives.bounds[0, 0])

            Y_best = np.max(Y_norm) if objectives.max_bool_dict[objectives.names[0]] else np.min(Y_norm)

            logger.info(f"Identifying approximate conditions with the highest expected improvement")
            with np.errstate(divide="warn"):
                sigma = Y_new_points_stddev_norm * self.exploration_factor
                if objectives.max_bool_dict[objectives.names[0]]:
                    Z = (Y_new_points_norm - Y_best) / sigma
                else:
                    Z = (Y_best - Y_new_points_norm) / sigma

                ei = sigma * (norm.pdf(Z) + (Z * norm.cdf(Z)))
                ei[sigma == 0.0] = 0.0

            sorting_indices = ei[:, -1].argsort()
            index_array = np.arange(X_new_points.shape[0]).reshape(-1, 1)
            concat = np.hstack((X_new_points, Y_new_points, ei))
            concat_sorted = concat[sorting_indices][::-1]
            index_array_sorted = index_array[sorting_indices][::-1]

            X_ei = concat_sorted[:, :-2]
            Y_ei = concat_sorted[:, -2].flatten()

            X_ei_norm = ((X_ei - np.array(variables.lower_bounds))) / (
                np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
            )

            # 0 index for the ei one at the top of the array
            chosen_indices = [0]
            # cosine similarity index of 1.0 for the ei one at the top of the array
            cosine_simularity_list = [1.0]
            # Comparing the X_values for the ei with every other entry down the sorted X_array
            for index, x in enumerate(X_ei_norm):
                cosine_simularity_index = 1 - spatial.distance.cosine(X_ei_norm[0], x)
                # Only passes if the cosine similarity index of a given x row is greater than 0.1 different from all of
                # the x's in the list
                if all((((cs - cosine_simularity_index) ** 2) ** 0.5) > 0.1 for cs in cosine_simularity_list):
                    cosine_simularity_list.append(cosine_simularity_index)
                    chosen_indices.append(index)

                if len(chosen_indices) == self.num_candidates:
                    break

            X_ei = X_ei[chosen_indices]
            Y_ei = Y_ei[chosen_indices].astype(np.float64)

            # Writes the indexes of the selected candidates in the sample pool
            if isinstance(self.sampler, PoolBased):
                self.sampler.index = index_array_sorted[: self.num_candidates].flatten().tolist()

                logger.info("Conditions were successfully selected using expected improvement")
                return X_ei, Y_ei

            elif isinstance(self.sampler, SampleGenerator):

                def fun(X_norm):
                    X_norm_reshape = X_norm.astype(np.float64).reshape(X_ei.shape)
                    X = (
                        X_norm_reshape * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                    ) + np.array(variables.lower_bounds)

                    if variables.num_cat_descriptor_var > 0:
                        X = variables.descriptor_to_name(X)

                    Y, Y_stddev = objectives.evaluate(X)

                    Y_norm = (Y - objectives.bounds[0, 0]) / (objectives.bounds[0, 1] - objectives.bounds[0, 0])
                    # I think this should be fine to minmax normalise like this
                    Y_stddev_norm = Y_stddev / (objectives.bounds[0, 1] - objectives.bounds[0, 0])

                    with np.errstate(divide="warn"):
                        sigma = Y_stddev_norm * self.exploration_factor
                        if objectives.max_bool_dict[objectives.names[0]]:
                            Z = (Y_norm - Y_best) / sigma
                        else:
                            Z = (Y_best - Y_norm) / sigma

                        ei = sigma * (norm.pdf(Z) + (Z * norm.cdf(Z)))
                        ei[sigma == 0.0] = 0.0

                    ei_sum = np.sum(ei)

                    return -1 * ei_sum

                # Convert the names of categorical variables to their descriptor values
                if variables.num_cat_descriptor_var > 0:
                    X_ei = variables.categorical_transform(X_ei)

                # Normalise all variable values against the bounds
                X_ei_norm = (X_ei - np.array(variables.lower_bounds)) / (
                    np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
                )

                logger.info(f"Optimising the variable location that provides the highest expected improvement")
                if constraints is None:
                    method = "L-BFGS-B"
                    consts = ()
                    options = {
                        # "iprint": 99,
                        "maxiter": 1000,
                        "maxfun": 1000000,
                    }
                else:
                    method = "SLSQP"
                    consts = constraints.create_scipy_constraints(self.num_candidates)
                    options = {
                        # "disp": True,
                        "maxiter": 1000
                    }

                np.random.seed(random.randint(0, 9999999))
                res = scipy_minimize(
                    fun,
                    X_ei_norm.flatten(),
                    method=method,
                    jac="3-point",
                    bounds=Bounds(
                        lb=np.zeros_like(X_ei_norm).flatten(),
                        ub=np.ones_like(X_ei_norm).flatten(),
                        keep_feasible=True,
                    ),
                    constraints=consts,
                    options=options,
                )

                if res.success == True:
                    res_X = res.x.astype(np.float64).reshape(X_ei.shape)

                    # Removes samples that are very similar and replaces them with the original values before
                    # optimization
                    selected_X = copy.copy(res_X)
                    for index_a, candidate_a in enumerate(selected_X):
                        for index_b, candidate_b in enumerate(res_X):
                            if index_a != index_b:
                                if np.all((np.isclose(candidate_a, candidate_b, rtol=1e-2))):
                                    selected_X[index_b] = X_ei_norm[index_b]

                    true_fun = fun(selected_X.flatten())

                    selected_X = (
                        selected_X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                    ) + np.array(variables.lower_bounds)

                    if variables.num_cat_var > 0:
                        selected_X = variables.categorical_values_euc(selected_X)
                    if variables.num_cat_descriptor_var > 0:
                        selected_X = variables.descriptor_to_name(selected_X)

                    selected_Y, Y_new_stddev = objectives.evaluate(selected_X)

                    success_res_fun_list.append(true_fun)
                    success_selected_X_list.append(selected_X)
                    success_selected_Y_list.append(selected_Y)

                else:
                    true_fun = res.fun

                    if variables.num_cat_var > 0:
                        X_ei = variables.categorical_values_euc(X_ei)
                    if variables.num_cat_descriptor_var > 0:
                        X_ei = variables.descriptor_to_name(X_ei)

                    fail_res_fun_list.append(true_fun)
                    fail_selected_X_list.append(X_ei)
                    fail_selected_Y_list.append(Y_ei.reshape(-1, 1) if Y_ei.ndim == 1 else Y_ei)

                print(
                    f'{iteration + 1}/{self.num_restarts}, EI optimisation: {"Success" if res.success else "Fail"}, '
                    f"iteration time (s) = {time.perf_counter() - start_time:>4.2f}, current function value: {true_fun}"
                    f", best function value: {min(success_res_fun_list + fail_res_fun_list)}"
                )

        if len(success_res_fun_list) > 0:
            logger.info("Successfully optimised the variable location that provides the highest expected improvement")
            best_fun_index = success_res_fun_list.index(min(success_res_fun_list))

            return success_selected_X_list[best_fun_index], success_selected_Y_list[best_fun_index]

        else:
            logger.warning(
                "Optimisation of the variable location that provides the highest expected improvement failed. Will use "
                "the best failed X-values"
            )
            best_fun_index = fail_res_fun_list.index(min(fail_res_fun_list))

            return fail_selected_X_list[best_fun_index], fail_selected_Y_list[best_fun_index]

    def ehvi(
        self,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected hypervolume improvement

        Parameters
        ----------
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected hypervolume improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected hypervolume improvement

        """
        logger.info(f"Identifying conditions with the highest expected hypervolume improvement")

        # Construct hypervolume reference point
        ref_point = self.build_ref_point(objectives.max_bool_dict)

        success_res_fun_list = []
        success_selected_X_list = []
        success_selected_Y_list = []
        fail_res_fun_list = []
        fail_selected_X_list = []
        fail_selected_Y_list = []
        for iteration in range(self.num_restarts):
            start_time = time.perf_counter()

            if isinstance(self.sampler, PoolBased):
                logger.info(f"Returning X values from the provided sample pool")
                X_new_points = self.sampler.X_pool
                index_combos = list(combinations(np.arange(X_new_points.shape[0]), self.num_candidates))
            elif isinstance(self.sampler, SampleGenerator):
                X_new_points = self.sampler.generate_samples(variables, constraints)

            Y_new_points, Y_new_points_stddev = objectives.evaluate(X_new_points)

            sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
            sign_adjusted_Y_new_points = self.Y_norm_minmax_transform(
                Y_new_points, objectives.bounds, objectives.max_bool_dict
            )

            # Min_max normalisation of Y standard deviation
            # I think this should be fine to minmax normalise like this
            for obj_index, _ in enumerate(objectives.bounds):
                Y_new_points_stddev[:, obj_index] = Y_new_points_stddev[:, obj_index] / (
                    objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
                )

            # Creates a list of expected hypervolume improvements for all combinations of the new Y points
            logger.info(f"Identifying approximate conditions with the highest expected hypervolume improvement")
            ehvi = ExpectedHypervolumeImprovement(
                ref_point=torch.tensor(ref_point, dtype=torch.double),
                Y=torch.tensor(sign_adjusted_Y, dtype=torch.double),
            )

            # Create every combination of the new possible samples and determines the ehvi of all of them
            X_new_points_combos = list(combinations(X_new_points, self.num_candidates))
            Y_new_points_combos = list(combinations(Y_new_points, self.num_candidates))
            Y_new_points_stddev_combos = list(combinations(Y_new_points_stddev, self.num_candidates))
            sign_adjusted_Y_new_points_combos = list(combinations(sign_adjusted_Y_new_points, self.num_candidates))

            def ehvi_process(candidates_Y, candidates_Y_stddev):
                try:
                    val = ehvi.ehvi_calc(
                        Y_new=candidates_Y,
                        Y_new_stddev=candidates_Y_stddev,
                        exploration_factor=self.exploration_factor,
                    )
                    return val
                except ValueError:
                    return torch.tensor(1e-8, dtype=torch.double)

            ehvi_list = Parallel(n_jobs=4)(
                delayed(ehvi_process)(
                    torch.tensor(np.array(candidates_Y), dtype=torch.double),
                    torch.tensor(np.array(candidates_Y_stddev), dtype=torch.double),
                )
                for candidates_Y, candidates_Y_stddev in zip(
                    sign_adjusted_Y_new_points_combos, Y_new_points_stddev_combos
                )
            )

            # Selects the variables and objective values of the candidate(s) that produced the highest expected
            # hypervolume improvement
            max_idx = ehvi_list.index(max(ehvi_list))
            X_ehvi = np.array(X_new_points_combos[max_idx])
            Y_ehvi = np.array(Y_new_points_combos[max_idx])

            # Writes the indexes of the selected candidates in the sample pool
            if isinstance(self.sampler, PoolBased):
                self.sampler.index = list(index_combos[max_idx])

                logger.info("Conditions were successfully selected using expected hypervolume improvement")
                return X_ehvi, Y_ehvi

            elif isinstance(self.sampler, SampleGenerator):
                # The objective function to be minimized for the scipy optimize minimize method
                def fun(X_norm):
                    X_norm_reshape = X_norm.astype(np.float64).reshape(X_ehvi.shape)
                    X = (
                        X_norm_reshape * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                    ) + np.array(variables.lower_bounds)

                    if variables.num_cat_descriptor_var > 0:
                        X = variables.descriptor_to_name(X)

                    Y, Y_stddev = objectives.evaluate(X)

                    sign_adjusted_Y = self.Y_norm_minmax_transform(Y, objectives.bounds, objectives.max_bool_dict)
                    # Min_max normalisation of Y standard deviation
                    # I think this should be fine to minmax normalise like this
                    for obj_index, _ in enumerate(objectives.bounds):
                        Y_stddev[:, obj_index] = Y_stddev[:, obj_index] / (
                            objectives.bounds[obj_index, 1] - objectives.bounds[obj_index, 0]
                        )

                    try:
                        fun_val = -1 * ehvi.ehvi_calc(
                            Y_new=torch.tensor(sign_adjusted_Y, dtype=torch.double),
                            Y_new_stddev=torch.tensor(Y_stddev, dtype=torch.double),
                            exploration_factor=self.exploration_factor,
                        )
                    except ValueError:
                        fun_val = -1 * torch.tensor(1e-8, dtype=torch.double)

                    return fun_val.item()

                # Convert the names of categorical variables with their descriptor values
                if variables.num_cat_descriptor_var > 0:
                    X_ehvi = variables.categorical_transform(X_ehvi)

                # Normalise all variable values against the bounds
                X_ehvi_norm = (X_ehvi - np.array(variables.lower_bounds)) / (
                    np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
                )

                # Performs the scipy optimize minimize method
                logger.info(
                    f"Optimising the variable location that provides the highest expected hypervolume improvement"
                )
                if constraints is None:
                    method = "L-BFGS-B"
                    consts = ()
                    options = {
                        # "iprint": 99,
                        "maxiter": 1000,
                        "maxfun": 1000000,
                    }
                else:
                    method = "SLSQP"
                    consts = constraints.create_scipy_constraints(self.num_candidates)
                    options = {
                        # "disp": True,
                        "maxiter": 1000
                    }

                np.random.seed(random.randint(0, 9999999))
                res = scipy_minimize(
                    fun,
                    X_ehvi_norm.flatten(),
                    method=method,
                    jac="3-point",
                    bounds=Bounds(
                        lb=np.zeros_like(X_ehvi_norm).flatten(),
                        ub=np.ones_like(X_ehvi_norm).flatten(),
                        keep_feasible=True,
                    ),
                    constraints=consts,
                    options=options,
                )

                if res.fun * -1 > max(ehvi_list).item():
                    if not res.success == True:
                        logger.warning(
                            f"Attempts to optimise the variable location that provides the highest expected "
                            f"hypervolume improvement failed to converge but still improved and so will use the new X "
                            f"values"
                        )

                    # Transform the optimised X value and calculate the corresponding objective Y values
                    res_X = res.x.astype(np.float64).reshape(X_ehvi.shape)

                    # Removes samples that are very similar and replaces them with the original values before
                    # optimization
                    selected_X = copy.copy(res_X)
                    for index_a, candidate_a in enumerate(selected_X):
                        for index_b, candidate_b in enumerate(res_X):
                            if index_a != index_b:
                                if np.all((np.isclose(candidate_a, candidate_b, rtol=1e-2))):
                                    selected_X[index_b] = X_ehvi_norm[index_b]

                    true_fun = fun(selected_X.flatten())

                    selected_X = (
                        selected_X * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
                    ) + np.array(variables.lower_bounds)

                    if variables.num_cat_var > 0:
                        selected_X = variables.categorical_values_euc(selected_X)
                    if variables.num_cat_descriptor_var > 0:
                        selected_X = variables.descriptor_to_name(selected_X)

                    selected_Y, selected_Y_stddev = objectives.evaluate(selected_X)

                    success_res_fun_list.append(true_fun)
                    success_selected_X_list.append(selected_X)
                    success_selected_Y_list.append(selected_Y)

                else:
                    true_fun = res.fun

                    if variables.num_cat_var > 0:
                        X_ehvi = variables.categorical_values_euc(X_ehvi)
                    if variables.num_cat_descriptor_var > 0:
                        X_ehvi = variables.descriptor_to_name(X_ehvi)

                    fail_res_fun_list.append(true_fun)
                    fail_selected_X_list.append(X_ehvi)
                    fail_selected_Y_list.append(Y_ehvi)

                print(
                    f'{iteration + 1}/{self.num_restarts}, EHVI optimisation: {"Success" if res.success else "Fail"}, '
                    f"iteration time (s) = {time.perf_counter() - start_time:>4.2f}, current function value: {true_fun}"
                    f", best function value: {min(success_res_fun_list + fail_res_fun_list)}"
                )

        if len(success_res_fun_list) > 0:
            logger.info("Successfully optimised the variable location that provides the highest expected improvement")
            best_fun_index = success_res_fun_list.index(min(success_res_fun_list))

            return success_selected_X_list[best_fun_index], success_selected_Y_list[best_fun_index]

        else:
            logger.warning(
                "Optimisation of the variable location that provides the highest expected improvement failed. Will "
                "use the best failed X-values"
            )
            best_fun_index = fail_res_fun_list.index(min(fail_res_fun_list))

            return fail_selected_X_list[best_fun_index], fail_selected_Y_list[best_fun_index]

    def qnei_botorch(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected improvement using the qNoisyExpectedImprovement
        acquisition function from the BoTorch package

        Parameters
        ----------
        X: np.ndarray
            An array containing X-values for all variables
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement

        """
        logger.info(f"Identifying conditions with the highest expected improvement")

        # BoTorch acquisition functions require the GP models for minimising objectives to be fitted using values that
        # have been multiplied by -1. This will re-fit the GP models with Y-values that have the correct sign
        temp_objectves = copy.copy(objectives)
        temp_objectves.objectives[0].fit_regressor(X, Y if objectives.objectives[0].obj_max_bool else -Y, variables)

        # Convert the names of categorical variables with their descriptor values
        if variables.num_cat_descriptor_var > 0:
            X = variables.categorical_transform(X)

        # Instantiate the single objective qNoisyExpectedImprovement acquisition function
        # This acquisition function is not using min-max normalised X-values
        acq_func = qNoisyExpectedImprovement(
            model=temp_objectves.objectives[0].obj_function.model,
            X_baseline=torch.tensor(X),
            sampler=SobolQMCNormalSampler(num_samples=1024),
        )

        # Optimises the acquisition function and returns the suggested candidates
        selected_X, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=torch.stack((torch.tensor(variables.lower_bounds), torch.tensor(variables.upper_bounds))),
            q=self.num_candidates,
            num_restarts=self.num_restarts,
            raw_samples=1024,
            equality_constraints=constraints.create_botorch_constraints("eq") if constraints is not None else None,
            inequality_constraints=constraints.create_botorch_constraints("ineq") if constraints is not None else None,
            options={"batch_limit": 5, "maxiter": 1000},
            sequential=True,
        )
        selected_X = selected_X.detach().cpu().numpy()

        if variables.num_cat_descriptor_var > 0:
            selected_X = variables.descriptor_to_name(selected_X.detach().cpu().numpy())

        selected_Y, _ = temp_objectves.evaluate(selected_X)

        return selected_X, selected_Y if objectives.objectives[0].obj_max_bool else -selected_Y

    def qnehvi_botorch(
        self,
        X: np.ndarray,
        Y: np.ndarray,
        variables: VariablesList,
        objectives: ObjectivesList,
        constraints: Optional[ConstraintsList] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """

        Function that determines the candidates with the best expected improvement using the
        qNoisyExpectedHypervolumeImprovement acquisition function from the BoTorch package

        Parameters
        ----------
        X: np.ndarray
            An array containing X-values for all variables
        Y: np.ndarray
            An array containing Y-values for all objectives
        variables: VariablesList
            VariablesList object that contains information about all variables
        objectives: ObjectivesList
            ObjectivesList object that contains information about all objectives
        constraints: ConstraintsList, Default = None
            ConstraintsList object that contains information about all constraints

        Returns
        -------
        selected_X: np.ndarray
            Array of X-values of the identified candidates' that have the highest expected improvement
        selected_Y: np.ndarray
            Array of Y-values of the identified candidates' that have the highest expected improvement

        """
        logger.info(f"Identifying conditions with the highest expected hypervolume improvement")

        # BoTorch acquisition functions require the GP models for minimising objectives to be fitted using values that
        # have been multiplied by -1. This will re-fit the GP models with Y-values that have the correct sign
        temp_objectves = copy.copy(objectives)
        for obj_index, obj in enumerate(temp_objectves.objectives):
            obj.fit_regressor(X, Y[:, obj_index] if obj.obj_max_bool else -Y[:, obj_index], variables)

        # Create the ModelListGP object for convenience
        model = ModelListGP(*[obj.obj_function.model for obj in temp_objectves.objectives])
        mll = SumMarginalLogLikelihood(model.likelihood, model)
        model.eval()
        mll.eval()

        # Convert the names of categorical variables with their descriptor values
        if variables.num_cat_descriptor_var > 0:
            X = variables.categorical_transform(X)

        # Normalise all variable values against the bounds
        X_norm = (X - np.array(variables.lower_bounds)) / (
            np.array(variables.upper_bounds) - np.array(variables.lower_bounds)
        )

        # Instantiate the multi-objective qNoisyExpectedHypervolumeImprovement acquisition function
        acq_func = qNoisyExpectedHypervolumeImprovement(
            model=model,
            ref_point=self.build_ref_point(temp_objectves.max_bool_dict),  # use known reference point
            X_baseline=torch.tensor(X_norm),
            prune_baseline=True,
            sampler=SobolQMCNormalSampler(num_samples=1024),
        )

        # Optimises the acquisition function and returns the suggested candidates
        selected_X_norm, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=torch.stack((torch.zeros(X_norm.shape[1]), torch.ones(X_norm.shape[1]))),
            q=self.num_candidates,
            num_restarts=self.num_restarts,
            raw_samples=1024,
            equality_constraints=constraints.create_botorch_constraints("eq") if constraints is not None else None,
            inequality_constraints=constraints.create_botorch_constraints("ineq") if constraints is not None else None,
            options={"batch_limit": 5, "maxiter": 1000},
            sequential=True,
        )

        # Un-min-max normalises the suggested candidates
        selected_X = (
            selected_X_norm.detach().cpu().numpy()
            * (np.array(variables.upper_bounds) - np.array(variables.lower_bounds))
        ) + np.array(variables.lower_bounds)

        if variables.num_cat_descriptor_var > 0:
            selected_X = variables.descriptor_to_name(selected_X)

        selected_Y, _ = temp_objectves.evaluate(selected_X)

        for obj_index, obj in enumerate(temp_objectves.objectives):
            if not obj.obj_max_bool:
                selected_Y[:, obj_index] = -selected_Y[:, obj_index]

        return selected_X, selected_Y



# This Expected Hypervolume Improvement acquisition function code was copied and adapted from the BoTorch package
# M. Balandat, B. Karrer, D. R. Jiang, S. Daulton, B. Letham, A. G. Wilson, and E. Bakshy. BoTorch: A Framework for
# Efficient Monte-Carlo Bayesian Optimization. Advances in Neural Information Processing Systems 33, 2020.
# http://arxiv.org/abs/1910.06403
# https://github.com/pytorch/botorch
# https://botorch.org/

from itertools import product

import torch
from botorch.utils.multi_objective.box_decompositions.non_dominated import FastNondominatedPartitioning
from torch import Tensor
from torch.distributions import Normal


class ExpectedHypervolumeImprovement:
    """

    Class for the Expected Hypervolume Improvement acquisition function adapted from the BoTorch package

    """

    def __init__(self, ref_point: Tensor, Y: Tensor):
        self.ref_point = ref_point
        self.partitioning = FastNondominatedPartitioning(
            ref_point=self.ref_point,
            Y=Y,
        )

        cell_bounds = self.partitioning.get_hypercell_bounds()
        self.cell_lower_bounds = cell_bounds[0]
        self.cell_upper_bounds = cell_bounds[1]

        self._cross_product_indices = torch.tensor(
            list(product(*[[0, 1] for _ in range(ref_point.shape[0])])),
            dtype=torch.long,
            device=ref_point.device,
        )

    def psi(self, lower: Tensor, upper: Tensor, mu: Tensor, sigma: Tensor) -> Tensor:
        u = (upper - mu) / sigma
        return sigma * Normal(0, 1).log_prob(u).exp() + (mu - lower) * (1 - Normal(0, 1).cdf(u))

    def nu(self, lower: Tensor, upper: Tensor, mu: Tensor, sigma: Tensor) -> Tensor:
        return (upper - lower) * (1 - Normal(0, 1).cdf((upper - mu) / sigma))

    def ehvi_calc(self, Y_new: Tensor, Y_new_stddev: Tensor, exploration_factor: int = 2) -> Tensor:
        # Reshaped to batch_size number of rows in the tensors
        Y_new = Y_new.reshape(Y_new.shape[0], 1, Y_new.shape[1])
        Y_new_stddev = Y_new_stddev.reshape(Y_new_stddev.shape[0], 1, Y_new_stddev.shape[1])

        # Promoting exploration
        Y_new_stddev = Y_new_stddev * exploration_factor

        cell_upper_bounds = self.cell_upper_bounds.clamp_max(1e10 if Y_new.dtype == torch.double else 1e8)

        psi_lu = self.psi(
            lower=self.cell_lower_bounds,
            upper=cell_upper_bounds,
            mu=Y_new,
            sigma=Y_new_stddev,
        )
        psi_ll = self.psi(
            lower=self.cell_lower_bounds,
            upper=self.cell_lower_bounds,
            mu=Y_new,
            sigma=Y_new_stddev,
        )

        nu = self.nu(
            lower=self.cell_lower_bounds,
            upper=cell_upper_bounds,
            mu=Y_new,
            sigma=Y_new_stddev,
        )

        psi_diff = psi_ll - psi_lu

        stacked_factors = torch.stack([psi_diff, nu], dim=-2)

        all_factors_up_to_last = stacked_factors.gather(
            dim=-2,
            index=self._cross_product_indices.expand(stacked_factors.shape[:-2] + self._cross_product_indices.shape),
        )

        return all_factors_up_to_last.prod(dim=-1).sum()



from typing import Tuple

import numpy as np
import sklearn


def remove_nan(arr1: np.ndarray, arr2: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """

    Removes slices in two arrays,arr1 and arr2, based on the indexes where there are NaN values in arr2. This method is
    useful to allow ML models to be fitted when you have an irregular number of Y-values for different objectives.
    Ignores any NaN values in arr1.

    Parameters
    ----------
    arr1: np.ndarray
        Array that will be manipulated based on the values in arr2
    arr2: np.ndarray
        The array with shape (n, ) that is used to identify the indexes where NaN values are. These indexes are applied
        as a boolean mask to arr1 and arr2

    Returns
    -------
    arr1: np.ndarray
        Array with slices removed that corresponded with the indexes of any NaN in the arr2 array
    arr2: np.ndarray
        Array with all NaN values removed

    """
    if arr1.shape[0] != arr2.shape[0]:
        raise ValueError(
            f"The arr1 shape at index 0 ({arr1.shape[0]}) is not the same as the arr2 shape at index 0 ({arr2.shape[0]})"
        )

    if arr2.ndim == 1:
        nan_mask = ~np.isnan(arr2)
    else:
        nan_mask = ~np.isnan(arr2).any(axis=1)

    return arr1[nan_mask], arr2[nan_mask]


def remove_all_nan_rows(Y: np.ndarray) -> np.ndarray:
    """

    Any rows that contain missing values (NaN) are removed. For example, useful for acquisition functions, which do not
    allow for irregular arrays of objective values.

    Parameters
    ----------
    Y: np.ndarray
        Y array to check for NaN values

    """
    if Y.ndim == 1:
        return Y[~np.isnan(Y)]
    else:
        return Y[~np.isnan(Y).any(axis=1)]


def sort(X: np.ndarray, Y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """

    Re-orders X and Y arrays such that the Y-values are in ascending order

    """
    if len(Y.shape) == 1:
        Y = Y.reshape(-1, 1)

    XY = np.hstack((X, Y))
    XY_sorted = XY[XY[:, X.shape[1]].argsort()]
    X_sorted = XY_sorted[:, :-1]
    Y_sorted = XY_sorted[:, -1].reshape(-1, 1)

    return X_sorted, Y_sorted.flatten()


def train_test_split(array: np.ndarray, splitnum: int) -> Tuple[np.ndarray, np.ndarray]:
    """

    Splits the provided array by taking out a row every user-defined number (splitnum) for the test set. The remaining
    data in the array are used in the training set

    Parameters
    ----------
    array: np.ndarray

    splitnum: int
        e.g. splitnum = 5 will cause every 5th row in the provided array to be extracted for the test set.

    Returns
    -------
    array_train: np.ndarray
        Training set created from the provided array
    array_test: np.ndarray
        Test set created from the provided array

    """
    array_test_indexes = [x for x in range(0, array.shape[0], splitnum)]
    array_train_indexes = list(set(list(range(array.shape[0]))) - set(array_test_indexes))

    array_train = array[array_train_indexes]
    array_test = array[array_test_indexes]

    return array_train, array_test


def train_val_test_split(array: np.ndarray, splitnum: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """

    Splits the provided array by taking out a row every user-defined number (splitnum) for the test set and
    (splitnum/2) for the validation set. The remaining data in the array are used in the training set

    Parameters
    ----------
    array: np.ndarray

    splitnum: int
        e.g. splitnum = 5 will cause every 5th row in the provided array to be extracted for the test set. Every 5th
        row starting from the splitnum/2 row will be extracted for the validation set

    Returns
    -------
    array_train: np.ndarray
        Training set created from the provided array
    array_val: np.ndarray
        Validation set created from the provided array
    array_test: np.ndarray
        Test set created from the provided array

    """
    array_test_indexes = [x for x in range(0, array.shape[0], splitnum)]
    array_val_indexes = [x for x in range(int(splitnum / 2), array.shape[0], splitnum)]
    array_train_indexes = list(set(list(range(array.shape[0]))) - set(array_test_indexes) - set(array_val_indexes))

    array_train = array[array_train_indexes]
    array_val = array[array_val_indexes]
    array_test = array[array_test_indexes]

    return array_train, array_val, array_test


def train_val_test_split_blocks(array: np.ndarray, test_ratio: float) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """

    Splits the inputted array into three equally sized parts for training, validation, and test sets

    """
    array_train, array_val, array_test = np.split(
        array,
        [
            int((1 - (2 * test_ratio)) * array.shape[0]),
            int((1 - test_ratio) * array.shape[0]),
        ],
    )

    return array_train, array_val, array_test


def shuffle(X: np.ndarray, Y: np.ndarray, seed: int) -> Tuple[np.ndarray, np.ndarray]:
    """

    Shuffles the X and Y arrays according to the seed integer provided

    """
    X, Y = sklearn.utils.shuffle(X, Y, random_state=seed)

    return X, Y


def sort_train_test_split_shuffle(
    X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, seed: int = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """

    Sequentially (i) sorts the inputted X and Y arrays, (ii) splits into training and test sets, and then (iii)
    shuffles the training and test sets

    Parameters
    ----------
    X: np.ndarray
        X array containing variables data
    Y: np.ndarray
        Y array containing objectives data with shape (y, )
    test_ratio: float, Default = 0.2
        The proportion of inputted X and Y arrays to be split into the test set
    seed: int, Default = None
        The seed used for shuffling the data sets to ensure that it is consistent
    Returns
    -------
    X_train: np.ndarray
        X array to be used as training data
    X_test: np.ndarray
        x array to be used as test data
    Y_train: np.ndarray
        Y array to be used as training data
    Y_test: np.ndarray
        Y array to be used as test data

    """
    splitnum = int(1 / test_ratio)

    X_sorted, Y_sorted = sort(X, Y)

    X_train, X_test = train_test_split(X_sorted, splitnum)
    Y_train, Y_test = train_test_split(Y_sorted, splitnum)

    if seed == None:
        seed = 1
    # np.random.seed(seed)  # setting the random seed
    X_train, Y_train = shuffle(X_train, Y_train, seed)
    X_test, Y_test = shuffle(X_test, Y_test, seed)

    return X_train, X_test, Y_train, Y_test


def sort_train_val_test_split_shuffle(
    X: np.ndarray, Y: np.ndarray, test_ratio: float = 0.2, seed: int = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """

    Sequentially (i) sorts the inputted X and Y arrays, (ii) splits into training, validation, and test sets, and then
    (iii) shuffles the training, validation, and test sets

    Parameters
    ----------
    X: np.ndarray
        X array containing variables data
    Y: np.ndarray
        Y array containing objectives data with shape (y, )
    test_ratio: float, Default = 0.2
        The proportion of inputted X and Y arrays to be split into the test set
    seed: int, Default = None
        The seed used for shuffling the data sets to ensure that it is consistent
    Returns
    -------
    X_train: np.ndarray
        X array to be used as training data
    X_val: np.ndarray
        x array to be used as validation data
    X_test: np.ndarray
        x array to be used as test data
    Y_train: np.ndarray
        Y array to be used as training data
    Y_val: np.ndarray
        Y array to be used as validation data
    Y_test: np.ndarray
        Y array to be used as test data

    """
    splitnum = int(1 / test_ratio)

    X_sorted, Y_sorted = sort(X, Y)

    X_train, X_val, X_test = train_val_test_split(X_sorted, splitnum)
    Y_train, Y_val, Y_test = train_val_test_split(Y_sorted, splitnum)

    if seed == None:
        seed = 1
    # np.random.seed(seed)  # setting the random seed
    X_train, Y_train = shuffle(X_train, Y_train, seed)
    X_val, Y_val = shuffle(X_val, Y_val, seed)
    X_test, Y_test = shuffle(X_test, Y_test, seed)

    return X_train, X_val, X_test, Y_train, Y_val, Y_test



import os
from attrs import define, field
from typing import Callable, List, Tuple, Union, Dict, Any, Optional

import numpy as np
import torch
from torch import Tensor

import nemo_bo.utils.logger as logging_nemo
from nemo_bo.opt.variables import CategoricalVariable, VariablesList

try:
    logging_nemo.logging_path
    logger = logging_nemo.logging_nemo_child(os.path.basename(__file__))
except AttributeError:
    logger = logging_nemo.logging_nemo_master(os.path.basename(__file__))


@define
class LinearConstraint:
    """

    Implements a linear constraint on the input variables for the optimisation. For example, linear constraints could
    be:

        a(F1) + b(F2) = 0
        a(F1) + b(F2) >= 0

    where a and b are coefficients for the features F1 and F2 respectively, and the right-hand side of the equation is
    0. Using an equals sign, =, would form a linear equality constraint whereas defining as 'greater than or equal to'
    forms an inequality constraint.

    By default in NEMO, all inequality constraints must be defined as 'greater than or equal to'. Note: This aligns
    with how constraints are defined in the scipy package, but is the opposite when compared to the pymoo package

    Parameters
    ----------
    constraint_type: str
        Defines whether the linear constraint should be an equality or inequality linear constraint by assigning as
        either 'eq' or 'ineq' respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to constrain, e.g. ['F1', 'F2'] for the example equation shown above
    coefficients: List[Union[int, float]]
        The coefficient for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [a, b] for the example equation shown above
    rhs: Union[int, float]
        The value to set for the right-hand side of the equation, e.g. 0 for the example equation shown above

    """

    constraint_type: str
    variables: VariablesList
    features: List[str]
    coefficients: List[Union[int, float]]
    rhs: Union[int, float]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.continuous_var_names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self) -> [Tuple[Tensor, Tensor, float]]:
        """

        Function that creates the constraint tuple that is used in the polytope sampler

        """
        return (torch.tensor(self.index), torch.tensor(self.coefficients), self.rhs)

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return np.sum(x[new_index] * self.coefficients) - self.rhs

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its
        # constraints
        if self.constraint_type == "ineq":
            return np.sum(x[:, self.index] * [-coeff for coeff in self.coefficients], axis=1) - (self.rhs * -1)
        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            return (np.sum(x[:, self.index] * [-coeff for coeff in self.coefficients], axis=1) - (self.rhs * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) >= round(float(self.rhs), 4)
        else:
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) == round(float(self.rhs), 4)

    def build_botorch_constraint(self) -> Tuple[Tensor, Tensor, float]:
        return (torch.tensor(self.index), torch.tensor(self.coefficients), float(self.rhs))


@define
class NonLinearPowerConstraint:
    """

    Implements a linear constraint on the input variables for the optimisation. For example, linear constraints could
    be:

        a(F1)^x + b(F2)^y = 0
        a(F1)^x + b(F2)^y >= 0

    where a and b are coefficients, and x and y are powers (exponents) for the features F1 and F2 respectively, and the
    right-hand side of the equation is 0. Using an equals sign, =, would form a linear equality constraint whereas
    defining as 'greater than or equal to' forms an inequality constraint.

    By default in NEMO, all inequality constraints must be defined as 'greater than or equal to'. Note: This aligns
    with how constraints are defined in the scipy package, but is the opposite when compared to the pymoo package

    Parameters
    ----------
    constraint_type: str
        Defines whether the linear constraint should be an equality or inequality linear constraint by assigning as
        either 'eq' or 'ineq' respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to constrain, e.g. ['F1', 'F2'] for the example equation shown above
    coefficients: List[Union[int, float]]
        The coefficient for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [a, b] for the example equation shown above
    powers: List[int]
        The power (exponent) for every features in the constraint. The values in the provided list must be given in
        the same order as the provided features to constrain, e.g. [x, y] for the example equation shown above
    rhs: Union[int, float]
        The value to set for the right-hand side of the equation, e.g. 0 for the example equation shown above

    """

    constraint_type: str
    variables: VariablesList
    features: List[str]
    coefficients: List[Union[int, float]]
    powers: List[int]
    rhs: Union[int, float]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.continuous_var_names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with non-linear constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            lhs_eval_list = []
            for i, c, p in zip(new_index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[i]) ** p) * c))

            return sum(lhs_eval_list) - self.rhs

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its
        # constraints
        lhs_eval_list = []
        if self.constraint_type == "ineq":
            for i, c, p in zip(self.index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[:, i]) ** p) * -c))

            return np.sum(lhs_eval_list, axis=0) - (self.rhs * -1)

        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            for i, c, p in zip(self.index, self.coefficients, self.powers):
                lhs_eval_list.append((((x[:, i]) ** p) * -c))

            return (np.sum(lhs_eval_list, axis=0) - (self.rhs * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) >= round(float(self.rhs), 4)
        else:
            return np.around(np.sum(X[:, self.index] * self.coefficients, axis=1), 4) == round(float(self.rhs), 4)


@define
class FunctionalConstraint:
    """

    Implements a constraint on the input variables for the optimisation that is calculated using a pass function. This
    allows any type of non-linear function to be used. The constraint can be utilised as either an equality or
    inequality constraint. For example:

        x[0] - sin(x[1]) = 0
        x[0] - sin(x[1]) >= 0

        def fun(x, **fun_kwargs):
            return x[0] - np.sin(x[1])

        or

        fun = lambda x: x[0] - np.sin(x[1])

    By default in NEMO, all inequality constraints must be defined as 'greater than or equal to'. Note: This aligns
    with how constraints are defined in the scipy package, but is the opposite when compared to the pymoo package

    Parameters
    ----------
    constraint_type: str
        Defines whether the linear constraint should be an equality or inequality linear constraint by assigning as
        either 'eq' or 'ineq' respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to constrain, e.g. ['F1', 'F2'] for the example equation shown above
    fun: Callable
        A callable that calculates the lhs of the constraint by taking the X-values of the constraint, and also any
        kwargs that are passed using the fun_kwargs
    rhs: Union[int, float]
        The value to set for the right-hand side of the equation, e.g. 0 for the example equation shown above
    fun_kwargs: Dict[str, Any], Deafult = None
        A dictionary of keyword arguments to use with the passed function, fun
    """

    constraint_type: str
    variables: VariablesList
    features: List[str]
    fun: Callable
    rhs: Union[int, float]
    fun_kwargs: Optional[Dict[str, Any]] = None
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.continuous_var_names.index(f) for f in self.features]

        if self.fun_kwargs is None:
            self.fun_kwargs = {}

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with functional constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return self.fun(x[:, new_index], **self.fun_kwargs) - self.rhs

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its
        # constraints
        lhs_eval_list = []
        if self.constraint_type == "ineq":
            return self.fun(x[:, self.index], **self.fun_kwargs) - (self.rhs * -1)

        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            return (self.fun(x[:, self.index], **self.fun_kwargs) - (self.rhs * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around(self.fun(X[:, self.index], **self.fun_kwargs), 4) >= round(float(self.rhs), 4)
        else:
            return np.around(self.fun(X[:, self.index], **self.fun_kwargs), 4) == round(float(self.rhs), 4)


@define
class StoichiometricConstraint:
    """

    Implements a constraint on the input variables for the optimisation for controlling the ratio between two
    variables. The ratio between the two variables can either be set to be equal to a particular value or the ratio
    between feature2 to feature 1 to be greater than a particular value

    Parameters
    ----------
    constraint_type: str
        Defines whether the stoichiometric ratio of feature2/feature1 constraint should be exactly the same as the
        defined ratio or if it can be any value greater than the ratio by assigning as either 'eq' or 'ineq'
        respectively
    variables: VariablesList
        VariablesList object that contains information about all variables
    feature1: str
        The name of one of the feature to consider in the constraint, e.g. 'Solvent'
    feature2: str
        The name of the other feature to consider in the constraint, e.g. 'Base'
    ratio: int | float
        The stoichiometric ratio of feature2/feature1

    """

    constraint_type: str
    variables: VariablesList
    feature1: str
    feature2: str
    ratio: Union[int, float]
    index: List[int] = field(init=False)

    def __attrs_post_init__(self):
        self.index1 = [self.variables.names.index(self.feature1)]
        self.index2 = [self.variables.names.index(self.feature2)]

    def build_polytope_sampler_constraint(self) -> [Tuple[Tensor, Tensor, float]]:
        """

        Function that creates the constraint tuple that is used in the polytope sampler

        """
        raise RuntimeError("The polytope sampler cannot be used with stoichiometric constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array
        num_candidates: int
            The number of candidates desired per iteration from the optimisation

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        # new_index = [i + add_to_index for i in self.index]
        #
        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return (x[:, self.index1 + add_to_index] / x[:, self.index2 + add_to_index]) - self.ratio

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its
        # constraints
        if self.constraint_type == "ineq":
            return (x[:, self.index1] / x[:, self.index2]) - (self.ratio * -1)
        else:
            # an equality constraint can be expressed in this manner by squaring the inequality expression
            # See https://pymoo.org/misc/constraints.html for more details
            return ((x[:, self.index1] / x[:, self.index2]) - (self.ratio * -1)) ** 2

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        if self.constraint_type == "ineq":
            return np.around((X[:, self.index1] / X[:, self.index2]), 4) >= round(float(self.rhs), 4)
        else:
            return np.around((X[:, self.index1] / X[:, self.index2]), 4) == round(float(self.rhs), 4)


@define
class MaxActiveFeaturesConstraint:
    """

    Implements a constraint that limits the number of variables/features that have values greater than 0. For example,
    if max_active = 2, the examples 1 and 2 below would pass the constraint whereas example 3 would fail the
    constraint:

        - Example with 4 variables in the X array
        - Only need to consider variables at index 0, 1, 2
        - Ensure that only 2 variables from indexes 0, 1, and 2 are above 0 (max_active = 2)

        1. X = [0.5, 0.7, 0.0, 1.0] -> 2 variables from indexes 0, 1, 2 are above 0 -> Pass
        2. X = [0.5, 0.0, 0.3, 1.0] -> 2 variables from indexes 0, 1, 2 are above 0 -> Pass
        3. X = [0.5, 0.7, 0.3, 1.0] -> 3 variables from indexes 0, 1, 2 are above 0 -> Fail

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    features: List[str]
        The name of the features to consider in the constraint, e.g. ['F1', 'F2']
    max_active: int
        The maximum number of features from the provided names that can be above 0

    """

    variables: VariablesList
    features: List[str]
    max_active: int
    index: List[int] = field(init=False)
    constraint_type: str = field(init=False)

    def __attrs_post_init__(self):
        self.index = [self.variables.continuous_var_names.index(f) for f in self.features]

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with max active number of features constraints")

    def build_scipy_constraint(self, add_to_index: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        new_index = [i + add_to_index for i in self.index]

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # The values in the x array that come from the scipy minimize function need to be normalised between 0 and 1
            # Returns a one where the values are above 0 and zero where they equal 0
            x = np.where(x > 0, 1, 0)

            # Evaluate the one-zero X array for the constraint
            return np.sum(x[new_index] * ([-1] * len(new_index))) - (self.max_active * -1)

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # The values in the x array that come from pymoo need to be normalised between 0 and 1
        # Multiply coefficients and rhs by -1 because pymoo uses the less than or equal to inequality operator for its
        # constraints
        X = np.where(x > 0, 1, 0)
        return np.sum(X[:, self.index], axis=1) - self.max_active

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        X = np.where(X > 0, 1, 0)
        return np.sum(X[:, self.index], axis=1) <= self.max_active


@define
class CategoricalConstraint:
    """

    Implements a constraint that prevents chosen categorical levels (e.g. chemicals) from two variables/features from
    being simultaneously selected. For example, the examples 1 and 2 below would pass the constraint whereas examples
    3 and 4 would fail the constraint:

        - Example where variables at indexes 0 and 1 are 'Solvent' and 'Base' respectively
        - In the 'Solvent' variable, the categorical level called 'Water' cannot be selected at the same time as
        categorical levels called 'Sodium' or 'Potassium' in the 'Base' variable

        1. X = ['Water', 'Triethylamine'] -> Pass
        2. X = ['Toluene', 'Sodium'] -> Pass
        3. X = ['Water', 'Sodium'] -> Fail
        4. X = ['Water', 'Potassium'] -> Fail

    Parameters
    ----------
    variables: VariablesList
        VariablesList object that contains information about all variables
    feature1: str
        The name of one of the feature to consider in the constraint, e.g. 'Solvent'
    feature2: str
        The name of the other feature to consider in the constraint, e.g. 'Base'
    categorical_levels1: List[str | int | float]
        The categorical levels in feature1 that cannot be simultaneously selected with categorical levels in feature2,
        e.g. ['Water']
    categorical_levels2: List[str | int | float]
        The categorical levels in feature2 that cannot be simultaneously selected with categorical levels in feature1,
        e.g. ['Sodium', 'Potassium']

    """

    variables: VariablesList
    feature1: str
    feature2: str
    categorical_levels1: List[Union[str, int, float]]
    categorical_levels2: List[Union[str, int, float]]
    index1: int = field(init=False)
    index2: int = field(init=False)

    def __attrs_post_init__(self):
        self.index1 = self.variables.names.index(self.feature1)
        self.index2 = self.variables.names.index(self.feature2)

        if not isinstance(self.variables.variables[self.index1], CategoricalVariable):
            raise TypeError(f"The variable {self.feature1} is not a categorical variable")

        if not isinstance(self.variables.variables[self.index2], CategoricalVariable):
            raise TypeError(f"The variable {self.feature2} is not a categorical variable")

    def build_polytope_sampler_constraint(self):
        raise RuntimeError("The polytope sampler cannot be used with categorical constraints")

    def build_scipy_constraint(self, add_to_index: int, num_candidates: int) -> Callable:
        """

        Creates the constraint function that is supplied into the scipy minimize function

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        constraint_fun: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """

        def constraint_fun(x: np.ndarray) -> np.ndarray:
            # Saves the original shape of the X array
            x_shape = x.shape

            # Reshapes the X array
            x = x.reshape(num_candidates, -1)

            # Un-normalise the variables
            x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
                self.variables.lower_bounds
            )

            # Converts the descriptor values generated by the scipy minimize function into the categorical level that
            # is the closest by Euclidean distance
            x = self.variables.descriptor_to_name(x)

            # Checks if the forbidden categorical levels are in the chosen features
            for row_index, (x1, x2) in enumerate(
                zip(x[:, self.index1 + add_to_index], x[:, self.index2 + add_to_index])
            ):
                if x1 in self.categorical_levels1:
                    x[row_index, self.index1 + add_to_index] = 1
                if x2 in self.categorical_levels2:
                    x[row_index, self.index2 + add_to_index] = 1

            # Re-shape the X array back to the original shape
            x = x.reshape(x_shape)

            # Evaluate the X array for the constraint
            return (x[{self.index1 + add_to_index}] * -1) + (x[{self.index2 + add_to_index}] * -1) - (-1)

        return constraint_fun

    def evaluate_pymoo_constraint(self, x: np.ndarray) -> np.ndarray:
        """

        Evaluate the X array for the constraint in pymoo

        Parameters
        ----------
        x: np.ndarray
            The X array to be evaluated. Can be more than one row of X-values

        Returns
        -------
        np.ndarray:
            The output value in the array for a given row in the supplied X array will be used by the pymoo package to
            evaluate whether the constraint was met


        """
        # Un-normalise the variables
        x = (x * (np.array(self.variables.upper_bounds) - np.array(self.variables.lower_bounds))) + np.array(
            self.variables.lower_bounds
        )

        # Converts the descriptor values generated by pymoo into the categorical level that is the closest by
        # Euclidean distance
        x = self.variables.descriptor_to_name(x)

        # Checks if the forbidden categorical levels are in the chosen features
        for row_index, (x1, x2) in enumerate(zip(x[:, self.index1], x[:, self.index2])):
            if x1 in self.categorical_levels1:
                x[row_index, self.index1] = 1
            if x2 in self.categorical_levels2:
                x[row_index, self.index2] = 1

        return np.sum(x[:, [self.index1, self.index2]], axis=1) - 1

    def bool_evaluate_constraint(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether the constraint is met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables in their categorical form (e.g. as
            categorical level names and not as descriptors for CategoricalVariableWithDescriptors)

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        # Checks if the forbidden categorical levels are in the chosen features
        for row_index, (x1, x2) in enumerate(zip(X[:, self.index1], X[:, self.index2])):
            if x1 in self.categorical_levels1:
                X[row_index, self.index1] = 1
            if x2 in self.categorical_levels2:
                X[row_index, self.index2] = 1

        return np.sum((X[:, [self.index1, self.index2]]), axis=1) <= 1


@define
class ConstraintsList:
    """

    Class that contains information and functions for all constraints

    Parameters
    ----------
    constraints: List[Union[LinearConstraint, NonLinearConstraint, MaxActiveFeaturesConstraint, CategoricalConstraint]]
        List of all constraints

    """

    constraints: List[
        Union[
            LinearConstraint,
            NonLinearPowerConstraint,
            FunctionalConstraint,
            StoichiometricConstraint,
            MaxActiveFeaturesConstraint,
            CategoricalConstraint,
        ]
    ]
    n_constr: int = field(init=False)
    variables: VariablesList = field(init=False)

    def __attrs_post_init__(self):
        self.n_constr = len(self.constraints)
        self.variables = self.constraints[0].variables

    def create_scipy_constraints(self, num_candidates: int) -> List[Tuple[Tensor, Tensor, float]]:
        """

        Creates the constraint functions that are supplied into the scipy minimize function for all constraints

        Parameters
        ----------
        add_to_index: int
            Offsets the position of the feature to account for its position when more than one candidate is desired per
            iteration from the optimisation. This is offset is required because multiple candidates are supplied to the
            scipy minimize function in a flattened array

        Returns
        -------
        scipy_constraints: Callable
            Function that can take an X array and is supplied to the scipy minimize function as a constraint

        """
        scipy_constraints = []
        for constr in self.constraints:
            for candidate_number in range(num_candidates):
                if not isinstance(constr, CategoricalConstraint):
                    add_to_index = candidate_number * constr.variables.n_var
                else:
                    add_to_index = candidate_number * len(constr.variables.variables)
                if (
                    isinstance(constr, LinearConstraint)
                    or isinstance(constr, NonLinearPowerConstraint)
                    or isinstance(constr, FunctionalConstraint)
                    or isinstance(constr, StoichiometricConstraint)
                ):
                    constraint = {
                        "type": constr.constraint_type,
                        "fun": constr.build_scipy_constraint(add_to_index, num_candidates),
                    }
                elif isinstance(constr, MaxActiveFeaturesConstraint):
                    constraint = {"type": "ineq", "fun": constr.build_scipy_constraint(add_to_index)}
                elif isinstance(constr, CategoricalConstraint):
                    constraint = {"type": "ineq", "fun": constr.build_scipy_constraint(add_to_index, num_candidates)}
                scipy_constraints.append(constraint)

        return scipy_constraints

    def bool_evaluate_constraints(self, X: np.ndarray) -> np.ndarray:
        """

        Evaluate whether all constraints are met and returns a boolean array

        Parameters
        ----------
        X: np.ndarray
            The X array to be evaluated that contains the categorical variables as their categorical level names and
            not as descriptors for CategoricalVariableWithDescriptors and/or as their discrete values for
            CategoricalVariableDiscreteValues

        Returns
        -------
        np.ndarray:
            A boolean is returned for all rows in the supplied X array that indicates whether the constraint was met

        """
        # Ensures that the X array is in its numerical form with descriptors if any CategoricalVariableWithDescriptors
        # or CategoricalVariableDiscreteValues variables are present.
        if self.variables.num_cat_var > 0:
            X = self.variables.categorical_transform(X)

        pass_fail = np.zeros((X.shape[0], self.n_constr))
        for constr_index, constr in enumerate(self.constraints):
            if not isinstance(constr, CategoricalConstraint):
                pass_fail[:, constr_index] = constr.bool_evaluate_constraint(X)
            else:
                # CategoricalConstraint requires all descriptors that are part of CategoricalVariableWithDescriptors
                # variables to be converted to their categorical level name
                pass_fail[:, constr_index] = constr.bool_evaluate_constraint(self.variables.descriptor_to_name(X))

        return np.all(pass_fail, axis=1)

    def create_botorch_constraints(self, constraint_type: str) -> List[Tuple[Tensor, Tensor, float]]:
        constraints_list = []
        for constr_index, constr in enumerate(self.constraints):
            if isinstance(constr, LinearConstraint):
                if constr.constraint_type == constraint_type:
                    constraints_list.append(constr.build_botorch_constraint())

        return constraints_list
